from calculation import Calculation, VaspError
from composition import Composition
from element import *
from entry import Entry
from experiment import *
from queue import Job, Task
from resources import *
from structure import Structure, Atom
from thermodynamics import *
from utils import *



from django.db import models
from getpass import getuser
from collections import defaultdict
import os
import json
from datetime import datetime
from ase.io.cif import read_cif, parse_cif
import time

from resources import Project
from structure import Structure
from experiment import ICSD, Reference, Journal, Author
from composition import Composition
from calculation import Calculation
from thermodynamics import *
from element import Element
from utils import *
from grid import Grid

class Entry(models.Model):
    '''
    An Entry is a structure submitted to the database. This entry can then be:
        - checked for uniqueness
        - relaxed and evaluated
        - checked for ferro vs anti-ferromagnetism (tbd)
        - compared to other entries at the same composition
    '''

    ### structure properties
    path = models.CharField(max_length=255, db_index=True, unique=True)
    source_file = models.CharField(max_length=255, blank=True)
    name = models.CharField(max_length=255, db_index=True, null=True)
    formula = models.CharField(max_length=255, db_index=True, null=True)

    ntypes = models.IntegerField(blank=True, null=True)
    natoms = models.IntegerField(blank=True, null=True)
    keywords = models.CharField(max_length=255, default='')
    label = models.CharField(max_length=20, null=True)
    holds = models.CharField(max_length=255, default='')
    log = models.TextField(default='')
    unique = models.BooleanField(default=False)
    stable = models.BooleanField(default=False)
    done = models.BooleanField(default=False)

    ## links
    composition = models.ForeignKey('Composition', null=True)
    element = models.ManyToManyField('Element', null=True)
    reference = models.ForeignKey('Reference', null=True)
    icsd = models.OneToOneField('ICSD', related_name='entry',
            null=True, blank=True)
    project = models.ForeignKey(Project, blank=True, null=True)

    class Meta:
        db_table = 'entries'

    def __str__(self):
        return '%s : %s' % (self.id, self.name)

    def save(self, *args, **kwargs):
        super(Entry, self).save(*args, **kwargs)
        if hasattr(self, 'new_elements'):
            for elt in self.new_elements:
                self.element.add(elt)
            del self.new_elements
        if hasattr(self, 'new_structures'):
            for struct in self.new_structures:
                self.structure_set.add(struct)
            del self.new_structures

    ###
    ###  Creation
    ###

    @staticmethod
    def create(source, keywords=[], project=None, **kwargs):
        source_file = os.path.abspath(source)
        path = os.path.dirname(source_file)
        if Entry.objects.filter(path=path).exists():
            return Entry.objects.get(path=path)
        entry = Entry(**kwargs)
        entry.source_file = source_file
        entry.path = os.path.dirname(source_file)

        ## Read in source file
        if source_file.endswith('.cif'):
            try:
                poscar = cif_to_poscar(source_file)
                f = open(entry.path+'/POSCAR','w')
                f.write(poscar)
                f.close()
            except Exception, err:
                print 'ERROR:', err
                print 'path:', source_file
                a = read(source_file)
                write_vasp(entry.path+'/POSCAR', a, vasp5=True, direct=True)
            structure = Structure.create(entry.path+'/POSCAR')
        else:
            structure = Structure.create(source_file)
        
        entry.add_structure(structure, label='source')
        entry.formula = comp_to_formula(structure.comp)
        entry.name = comp_to_name(structure.comp, special='reduce')
        entry.assign_composition()
        entry.assign_elements()
        entry.ntypes = structure.ntypes
        entry.natoms = len(structure.atoms)

        for k in keywords:
            entry.add_keyword(k)
        if project is not None:
            entry.set_project(project)
        if source_file.endswith('.cif'):
            entry.read_addtl_cif_data()
        entry.write_log('created %s' % datetime.now())
        return entry

    ## helpers for filling in fields
    def assign_composition(self):
        composition = Composition.get(self.comp)
        self.composition = composition

    def assign_elements(self):
        for k in self.comp.keys():
            elt = Element.get(k)
            self.add_element(elt)

    def add_element(self, elt):
        if not hasattr(self, 'new_elements'):
            self.new_elements = []
        self.new_elements.append(elt)

    def read_addtl_cif_data(self):
        if not self.source_file.endswith('.cif'):
            return

        cif_data = parse_cif(self.source_file)[0][1]
        journal, created =\
                Journal.objects.get_or_create(code=cif_data['_citation_journal_id_astm'][0])
        if created:
            journal.full_name = cif_data['_citation_journal_full'][0]

        reference, created= Reference.objects.get_or_create(
                journal=journal,
                year=cif_data['_citation_year'][0],
                title=cif_data['_publ_section_title'])
        self.reference = reference
        
        for author in cif_data['_publ_author_name']:
            auth, created = Author.objects.get_or_create(name=author)
            reference.authors.add(auth)

        icsd = ICSD.parse_cif(self.source_file)
        self.icsd = icsd

        if any([s < 0.98 for s in cif_data['_atom_site_occupancy']]):
            print cif_data['_atom_site_occupancy']
            self.add_hold('partial')
        nom_comp = cif_data['_chemical_formula_sum']
        nom_dict = name_to_comp(nom_comp, special='unit')
        if set(nom_dict.keys()) != set(self.comp.keys()):
            self.add_hold('incomplete_composition')
        else:
            scales = [ self.unit_comp[k]/v for k,v in nom_dict.items() ]
            if any([ abs(1-s) > 0.05 for s in scales ]):
                self.add_hold('composition_mismatch_in_cif')
            elif len(set(scales)) > 1:
                self.add_keyword('slight_composition_mismatch')

    ###
    ### Dynamically calculated properties
    ###

    @property
    def space(self):
        return set([ e.symbol for e in self.elements])

    @property
    def elements(self):
        return list(self.element.all())

    @property
    def tasks(self):
        return list(self.task_set.all())

    @property
    def jobs(self):
        return list(self.job_set.all())

    @property
    def comp(self):
        return name_to_comp(self.formula)

    @property
    def nom_comp(self):
        return name_to_comp(self.formula, special='reduce')

    @property
    def unit_comp(self):
        return name_to_comp(self.formula, special='unit')

    @property
    def latex(self):
        return comp_to_latex(self.comp)

    @property
    def calculations(self):
        return list(self.calculation_set.all())

    @property
    def energy(self):
        if self.calculation_set.filter(settings='standard', done=True).exists():
            return self.calculation_set.get(settings='standard', done=True).energy_pa 
        else:
            return None

    @property
    def mass(self):
        return sum( Element.get(elt)*amt for elt, amt in self.unit_comp)

    @property
    def volume(self):
        if not self.relaxed is None:
            return self.relaxed.volume/self.natoms
        else:
            return self.input.volume/self.natoms

    ### holds
    def add_hold(self, hold):
        holds = self.get_holds()
        if hold not in holds:
            holds.append(hold)
        self.holds = ' '.join(holds)

    def get_holds(self):
        holds = self.holds.strip().split(' ')
        return [ h for h in holds if h ]

    def remove_hold(self, hold):
        holds = self.get_holds()
        holds.remove(hold)
        self.holds = ' '.join(holds)

    ## keywords
    def add_keyword(self, keyword):
        keywords = self.get_keywords()
        if keyword not in keywords:
            keywords.append(keyword)
        self.keywords = ' '.join(keywords)

    def get_keywords(self):
        return self.keywords.strip().split(' ')

    def remove_keyword(self, keyword):
        keywords = self.get_keywords()
        keywords.remove(keyword)
        self.keywords = ' '.join(keywords)

    ## log
    def write_log(self, event):
        self.log += '\n'
        self.log += str(event)

    ## project
    def set_project(self, project):
        proj, created = Project.objects.get_or_create(name=project)
        if created:
            print 'Making new project', project
            print 'NOTE: Has no available resources!'
        self.project = proj

    ## structures
    def add_structure(self, structure, label=None):
        if not hasattr(self, 'new_structures'):
            self.new_structures = []
        if label is not None:
            structure.label = label
        self.new_structures.append(structure)

    def get_structure(self, label):
        ret = None
        if hasattr(self, 'new_structures'):
            for struct in self.new_structures:
                if struct.label == label:
                    ret = struct
        test = self.structure_set.filter(label=label)
        if test.exists():
            return test[0]

    @property
    def structures(self):
        ret_structs = []
        for struct in self.structure_set.all():
            ret_structs.append(struct)
        if hasattr(self, 'new_structures'):
            for struct in self.new_structures:
                ret_structs.append(struct)
        return ret_structs

    @property
    def input(self):
        return self.get_structure('source')

    @property
    def coarse_relaxed(self):
        return self.get_structure('coarse_relaxed')

    @property
    def relaxed(self):
        return self.get_structure('relaxed')

    ## calculations

    def get_calculation(self, settings, **kwargs):
        test = self.calculation_set.filter(settings=settings, **kwargs)
        if test.count() == 1:
            return test[0]
        elif test.count() > 1:
            return test.order_by('energy')[0]
        elif test.count() == 0:
            return None

    def is_done(self, settings):
        calc = self.get_calculation(settings, done=True)
        if calc is None:
            return False
        else:
            return True

    def initialize(self, args=[]):
        if self.is_done('initialize'):
            return self.get_calculation('initialize', done=True)

        self.input.set_magnetism('ferro')
        calc = Calculation.do(self.input, entry=self,
                type='initialize', path=self.path+'/initialize', args=args)
        calc.npar = 2
        calc.write()
        if calc.done:
            if calc.magmom > 0.1:
                self.add_keyword('magnetic')
        calc.save()
        return calc

    def coarse_relax(self, args=[]):
        if self.is_done('coarse_relax'):
            return self.get_calculation('coarse_relax', done=True)

        calc = self.initialize(args=args)
        if not calc.done:
            calc.npar = 2
            calc.write()
            return calc
        if not 'magnetic' in self.keywords:
            self.input.set_magnetism('none')
        calc = Calculation.do(self.input, entry=self,
            type='coarse_relax', path=self.path+'/coarse_relax', args=args)
        if calc.done:
            if calc.output is None:
                calc.output = calc.input
                calc.save()
            calc.output.save()
            self.add_structure(calc.output, label='coarse_relaxed')
        elif not os.path.exists(self.path+'/coarse_relax/CHGCAR.gz'):
            os.system('cp %s/initialize/CHGCAR.gz %s/coarse_relax/' % 
                    (self.path, self.path))
        calc.save()
        self.save()
        return calc

    def fine_relax(self, args=[]):
        if self.is_done('fine_relax'):
            return self.get_calculation('fine_relax', done=True)

        calc = self.coarse_relax(args=args)
        if not calc.done:
            calc.npar = 2
            calc.write()
            return calc
        if 'magnetic' in self.keywords:
            self.coarse_relaxed.set_magnetism('ferro')

        calc = Calculation.do(self.coarse_relaxed, entry=self,
            type='fine_relax', path=self.path+'/fine_relax', args=args)
        if calc.done:
            if calc.output is None:
                calc.output = calc.input
                calc.save()
            self.add_structure(calc.output, label='relaxed')
        elif not os.path.exists(self.path+'/standard/CHGCAR.gz'):
            os.system('cp %s/coarse_relax/CHGCAR.gz %s/fine_relax/'
                    % (self.path, self.path))

        calc.save()
        self.save()
        return calc

    def standard(self, args=[]):
        if self.is_done('standard'):
            return self.get_calculation('standard', done=True)
        calc = self.fine_relax(args=args)
        if not calc.done:
            calc.npar = 2
            calc.write()
            return calc

        if 'magnetic' in self.keywords:
            self.relaxed.set_magnetism('ferro')

        calc = Calculation.do(self.relaxed, entry=self,
            type='standard', path=self.path+'/standard', args=args)
        if calc.done:
            #from thermopy import PhaseSpace
            Formation.get(calc, fit='standard')
            #s = PhaseSpace('-'.join(self.comp.keys()))
            #s.load('oqmd')
            #s.get_meta_stabilities()
            # maybe add to global phase diagram?
        elif not os.path.exists(self.path+'/standard/CHGCAR.gz'):
            os.system('cp %s/fine_relax/CHGCAR.gz %s/standard' % 
                    (self.path, self.path))
        calc.save()
        return calc

    @property
    def oqmd(self):
        calcs = self.calculation_set.filter(settings='standard', done=True)
        if calcs.exists():
            return calcs.order_by('energy')[0]
        else:
            return None

    def do(self, module, args=[]):
        return getattr(self, module)(args=args)

    @property
    def errors(self):
        return dict( ( c.path, c.get_errors()) for c in self.calculations )

    def move(self, path):
        path = os.path.abspath(path)
        try:
            os.system('mv %s %s' % (self.path, path))
        except Exception, err:
            print err
            return
        self.path = path
        print 'Moved %s to %s' % (self, path)
        self.save()

    @property
    def chg(self):
        if not hasattr(self, '_chg'):
            if not self.done:
                self._chg = False
            else:
                self._chg = Grid.load_xdensity(self.path+'/standard/CHGCAR.gz')
        return self._chg

    def reset(self):
        for structure in self.structures:
            if structure.label != 'source':
                structure.delete()
        
        for calc in self.calculations:
            calc.clean_start()
            calc.delete()

        for task in self.tasks:
            task.state = 0 
            task.save()

        for job in self.job_set.filter(state=1):
            job.collect()
            job.delete()
        self.job_set.all().delete()

        for dir in ['initialize', 'coarse_relax', 'fine_relax', 'standard']:
            os.system('rm -rf %s/%s &> /dev/null' % (self.path, dir))

    def visualize(self, structure='source'):
        os.system('/usr/local/bin/VESTA %s/POSCAR' % self.path)



#!/usr/bin/python
import itertools
import numpy as np
from numpy.linalg import norm
import re
import gzip
import subprocess
from collections import defaultdict

def gcd(a,b):
    '''
    Returns greatest common denominator of two numbers.
    
    Example:
    >>> gcd(4, 6)
    2
    '''
    while b:
        a,b=b,a%b
    return a

def get_gcd(numbers):
    '''
    Returns greatest common denominator of an iterator of numbers.

    Example:
    >>> get_gcd(12, 20, 32)
    4
    '''
    if len(numbers) == 1:
        return numbers[0]
    a=numbers.pop()
    b=numbers.pop()
    tmp_gcd = gcd(a,b)
    while numbers:
        tmp_gcd = gcd(tmp_gcd,numbers.pop())
    return tmp_gcd

def lcm(a,b):
    '''
    Returns least common multiple of two numbers.

    Example:
    >>> lcm(20,10)
    20
    '''
    return a*b/gcd(a,b)

def ffloat(string):
    '''
    In case of fortran digits overflowing and returing ********* this
    fuction will replace such values with 1e9
    '''

    try:
        new_float = float(string)
    except ValueError:
        if '*******' in string:
            new_float = 1e9
        else:
            print string
    return new_float

def angle(x, y):
    '''Return the angle between two lattice vectors

    Example:

    >>> angle([1,0,0], [0,1,0])
    90.0
    '''
    return np.arccos(np.dot(x,y)/(norm(x)*norm(y)))*180./np.pi

def basis_to_latparams(basis):
    """Returns the lattice parameters [a, b, c, alpha, beta, gamma].
    
    Example:
        
    >>> basis_to_latparams([[3,0,0],[0,3,0],[0,0,5]])
    [3, 3, 5, 90, 90, 90]
    """

    va, vb, vc = basis
    a = np.linalg.norm(va)
    b = np.linalg.norm(vb)
    c = np.linalg.norm(vc)
    alpha = angle(vb, vc)
    beta = angle(vc, va)
    gamma = angle(va, vb)
    return [a, b, c, alpha, beta, gamma]


def latparams_to_basis(latparam):
    '''Returns a bsis from the lattice parameters.

    Example:
    >>> latparams_to_basis([3, 3, 5, 90, 90, 90])
    [[3,0,0],[0,3,0],[0,0,5]]
    '''

    basis = []
    a, b, c = latparam[0:3]
    alpha, beta, gamma = latparam[3:6]
    basis.append([a, 0, 0])
    basis.append([b*np.cos(gamma), b*np.sin(gamma), 0 ])
    cx = np.cos(beta)
    cy = (np.cos(alpha) - np.cos(beta)*np.cos(gamma))/np.sin(gamma)
    cz = np.sqrt(1. - cx*cx - cy*cy)
    basis.append([c*cx, c*cy, c*cz])
    return np.array(basis)


def coord_to_bin(coord):
    return coords[0]

def coord_to_gtri(coord):
    '''( A+B/2, B*3^(1/2)/2 )'''
    return (coord[0] + coord[1]/2., 
            coord[1]*np.sqrt(3)/2)

def coord_to_gtet(coord):
    '''( A/2+B+C/2, A*3^(1/2)/2 + C*3^(1/2)/6, C*(2/3)^(1/2) )'''
    return (coord[0]/2 + coord[1] + coord[2]/2,
        coord[0]*np.sqrt(3)/2 + coord[2]*np.sqrt(3)/6.,
        coord[2]*np.sqrt(2./3))



from my_math import *
from my_strings import *
#import my_strings
from miedema import get_energy
from cif_to_poscar import *

def assign_type(string):
    res = string
    try:
        res = float(string)
    except ValueError:
        pass
    try:
        res = int(string)
    except ValueError:
        pass
    if string.lower() == 'true':
        res = True
    elif string.lower() == 'false':
        res = False
    elif string.lower() == 'none':
        res = None
    return res

def combinations_with_replacement(iterable, r):
    # combinations_with_replacement('ABC', 2) --> AA AB AC BB BC CC
    pool = tuple(iterable)
    n = len(pool)
    if not n and r:
        return
    indices = [0] * r
    yield tuple(pool[i] for i in indices)
    while True:
        for i in reversed(range(r)):
            if indices[i] != n - 1:
                break
        else:
            return
        indices[i:] = [indices[i] + 1] * (r - i)
        yield tuple(pool[i] for i in indices)



from qmpy.structure import Structure
from ase import Atoms

def to_ase(structure):
    atoms = Atoms(cell=structure.cell,
            scaled_positions=structure.scaled_coords, 
            magmoms=structure.magmoms,
            symbols=''.join(sorted(a.symbol for a in structure.atoms)))
    return atoms

def from_ase(atoms):
    struct = Structure()
    struct.cell = atoms.get_cell()
    for a in atoms: 
        atom = Atom()
        atom.symbol = a.symbol
        atom.x, atom.y, atom.z = a.position
        atom.direct = False
        struct.add_atom(atom)
    return struct



#!/usr/bin/python

import subprocess
import tempfile
import os
from collections import defaultdict
from ase.io.cif import parse_cif
from numpy import argsort,array,dot
from numpy.linalg import inv

ochoice=[48,50,59,68,70,85,86,88,125,126,129,130,133,134,
        137,138,141,142,201,203,222,224,227,228]
uchoice=[3,4,5,6,7,8,9,10,11,12,13,14,15]
hchoice=[146,148,155,160,161,166,167]

def cif_to_poscar(cif):
    cif_data = parse_cif(cif)[0][1]
    a=cif_data['_cell_length_a']
    b=cif_data['_cell_length_b']
    c=cif_data['_cell_length_c']
    alpha = cif_data['_cell_angle_alpha']
    beta = cif_data['_cell_angle_beta']
    gamma = cif_data['_cell_angle_gamma']
    sg = cif_data['_symmetry_int_tables_number']
    sg_name = cif_data['_symmetry_space_group_name_h-m']

    origin=False
    if int(sg) in ochoice and sg_name.endswith('Z'):
        origin=2

    wyksites={}
    for i in range(len(cif_data['_atom_site_label'])):
        label = cif_data['_atom_site_label'][i]
        elt = ''.join(c for c in label if c.isalpha)
        wyksites[label] = {'elt':elt}
        wyksites[label]['multiple'] =\
                cif_data['_atom_site_symmetry_multiplicity'][i]

        wyksites[label]['pos'] = [
                cif_data['_atom_site_fract_x'][i],
                cif_data['_atom_site_fract_y'][i],
                cif_data['_atom_site_fract_z'][i]]

    tmp = tempfile.mkdtemp(dir='/dev/shm')
    cwd = os.getcwd()
    os.chdir(tmp)

    # Create gulp.in
    gulp_instr = 'single\ntitle\n'
    gulp_instr += 'filler\n'
    gulp_instr += 'end\ncell\n'
    gulp_instr += '%.8f %.8f %.8f &\n %.8f %.8f %.8f\n' % (
            a, b, c, alpha, beta, gamma)
    gulp_instr += 'frac\n'
    for site in wyksites.values():
        sstr = ' '.join([ str(coord) for coord in site['pos'] ])
        gulp_instr += '%s %s\n' % (site['elt'], sstr)
    gulp_instr += 'space\n'
    gulp_instr += '%s\n' % sg_name.rstrip('HSZR')
    if origin:
        gulp_instr += 'origin 2\n'
    target = tempfile.mktemp(dir=tmp, suffix='.xyz')
    gulp_instr += 'output xyz '+target

    # Run GULP
    p1 = subprocess.Popen(['echo',gulp_instr],
            stdout=subprocess.PIPE)
    p2 = subprocess.Popen('gulp40',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=p1.stdout)
    p2.wait()
    out, err = p2.communicate()

    if err:
        gulp_instr = 'single\ntitle\n'
        gulp_instr += 'filler\n'
        gulp_instr += 'end\ncell\n'
        gulp_instr += '%.8f %.8f %.8f &\n %.8f %.8f %.8f\n' % (
                a, b, c, alpha, beta, gamma)
        gulp_instr += 'frac\n'
        for site in wyksites.values():
            sstr = ' '.join([ str(coord) for coord in site['pos'] ])
            gulp_instr += '%s %s\n' % (site['elt'], sstr)
        gulp_instr += 'space\n'
        gulp_instr += '%s\n' % sg
        if origin:
            gulp_instr += 'origin 2\n'
        target = tempfile.mktemp(dir=tmp, suffix='.xyz')
        gulp_instr += 'output xyz '+target
        # Run GULP
        p1 = subprocess.Popen(['echo',gulp_instr],
                stdout=subprocess.PIPE)
        p2 = subprocess.Popen('gulp40',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=p1.stdout)
        p2.wait()
        out, err = p2.communicate()

    if err:
        raise Exception

    stdout = out.split('\n')
    for n, line in enumerate(stdout):
        if 'Cartesian' in line:
            latvecs = '\n'.join(stdout[n+2:n+5])

    os.chdir(cwd)
    coords = defaultdict(list)
    atom_data = open(target)
    for line in atom_data:
        line = line.split()
        if len(line) != 4:
            continue
        atom = [float(x) for x in line[1:]]
        elt = line[0]
        if elt in ['D', 'T']:
            elt = 'H'
        dup = False
        for coord in coords[elt]:
            if all([ abs(x1-x2) < 0.1 for x1, x2 in zip(coord, atom)]):
                dup = True
        if not dup:
            coords[elt].append(atom)

    ordered_keys = sorted(coords)

    outstr = ' '.join(ordered_keys)+"\n"
    outstr += "1.0\n"
    outstr += latvecs+"\n"
    outstr += ' '.join(ordered_keys)+'\n'
    outstr += ' '.join(str(len(coords[k])) for k in ordered_keys)+'\n'
    outstr += "Cartesian\n"
    for k in ordered_keys:
        for atom in coords[k]:
            outstr += '%.8f %.8f %.8f\n' % (atom[0], atom[1], atom[2])
    outstr.rstrip('\n')
    return outstr



#!/usr/bin/python
import numpy as np
import time
import sys
# miedema.py v1.6 12-13-2012 Jeff Doak jeff.w.doak@gmail.com

def input_elements(filename):
    """Returns a dictionary containing as keys all the elements which have
    tabulated Miedema model input parameters. The value associated with each key
    is a list containing these input parameters. The parameters in the list are:
    Phi, n_ws^(1/3.), vol^(2/3.), Z, Valence, TM, RtoP, Htrans(kJ/mol)"""
    file = open(filename,"r")
    elements = {}
    for line in file:
        if line.startswith("#"):
            continue
        line = line.split()
        if len(line) == 1:
            elements[line[0]] = []
        if len(line) > 1:
            # Miedema File Format:
            # Element_name Phi Rho Vol Z Valence TM? RtoP Htrans
            # 0            1   2   3   4 5       6   7    8
            #              0   1   2   3 4       5   6    7
            # str          float...... int.......... float......
            key = line[0]
            val = []
            val.append(float(line[1]))
            val.append(float(line[2]))
            val.append(float(line[3]))
            val.append(int(line[4]))
            val.append(int(line[5]))
            val.append(int(line[6]))
            val.append(float(line[7]))
            val.append(float(line[8]))
            elements[key] = val
    file.close()
    return elements

def pick_P(elementA,elementB,database):
    """Chooses a value of P based on the transition metal status of the elements
    A and B. There are 3 values of P for the cases where both A and B are TM,
    where only one of A and B is a TM, and where neither are TMs."""
    possibleP = [14.2,12.35,10.7]
    if (database[elementA][5] + database[elementB][5]) == 2:
        # Both elementA and elementB are Transition Metals.
        return possibleP[0]
    elif (database[elementA][5] + database[elementB][5]) == 1:
        # Only one of elementA and elementB are Transition Metals.
        return possibleP[1]
    else:
        # Neither elementA nor elementB are Transition Metals.
        return possibleP[2]

def pick_RtoP(elementA,elementB,database):
    """Calculate and return the value of RtoP based on the transition metal
    status of elements A and B, and the elemental values of RtoP for elements A
    and B."""
    # List of Transition Metals as given in Fig 2.28 of 
    # de Boer, et al., Cohesion in Metals (1988) (page 66).
    tmrange = []
    tmrange.extend(range(20,30))
    tmrange.extend(range(38,48))
    tmrange.extend(range(56,58))
    tmrange.extend(range(72,80))
    tmrange.extend([90,92,94])
    # List of Non-Transition Metals as given in Fig 2.28 of 
    # de Boer, et al., Cohesion in Metals (1988) (page 66).
    nontmrange = []
    nontmrange.extend(range(3,8))
    nontmrange.extend(range(11,16))
    nontmrange.extend([19])
    nontmrange.extend(range(30,34))
    nontmrange.extend([37])
    nontmrange.extend(range(48,52))
    nontmrange.extend([55])
    nontmrange.extend(range(80,84))
    # If one of A,B is in tmrange and the other is in nontmrange, set RtoP
    # to the product of elemental values, otherwise set RtoP to zero.
    if (database[elementA][3] in tmrange) and (database[elementB][3] in
            nontmrange):
        RtoP = database[elementA][6]*database[elementB][6]
    elif (database[elementA][3] in nontmrange) and (database[elementB][3] in
            tmrange):
        RtoP = database[elementA][6]*database[elementB][6]
    else:
        RtoP = 0.0
    return RtoP

def pick_a(elementA,database):
    """Choose a value of a based on the valence of element A."""
    possible_a = [0.14,0.1,0.07,0.04]
    if database[elementA][4] == 1:
        return possible_a[0]
    elif database[elementA][4] == 2:
        return possible_a[1]
    elif database[elementA][4] == 3:
        return possible_a[2]
    #elif elementA in ["Ag","Au","Ir","Os","Pd","Pt","Rh","Ru"]:
    elif elementA in ["Ag","Au","Cu"]:
        return possible_a[2]
    else:
        return possible_a[3]

def gamma_ab(elementA,elementB,database):
    """Calculate and return the value of Gamma_AB (= Gamma_BA) for the solvation
    of element A in element B."""
    QtoP = 9.4  # Constant from Miedema's Model.
    phi = [database[elementA][0],database[elementB][0]]
    rho = [database[elementA][1],database[elementB][1]]
    delta_phi = phi[0] - phi[1]
    delta_rho = rho[0] - rho[1]
    average_rho = (1/rho[0] + 1/rho[1])/2.
    P = pick_P(elementA,elementB,database)
    RtoP = pick_RtoP(elementA,elementB,database)
    Gamma = P*(QtoP*delta_rho**2 - delta_phi**2 - RtoP)/average_rho
    Gamma = int(round(Gamma))
    return Gamma

def H_form_ord(elementA,elementB,xB,database):
    """Calculate the enthalpy of formation for an ordered compound of elements A
    and B with a composition xB of element B."""
    Gamma = gamma_ab(elementA,elementB,database)
    vol0_A = database[elementA][2]
    vol0_B = database[elementB][2]
    phi = [database[elementA][0],database[elementB][0]]
    htrans = [database[elementA][7],database[elementB][7]]
    # Determine volume scale parameter a.
    a_A = pick_a(elementA,database)
    a_B = pick_a(elementB,database)
    # Calculate surface concentrations using original volumes.
    c_S_A = (1-xB)*vol0_A/((1-xB)*vol0_A+xB*vol0_B)
    c_S_B = xB*vol0_B/((1-xB)*vol0_A+xB*vol0_B)
    # Calculate surface fractions for ordered compounds using original volumes.
    f_BA = c_S_B*(1+8*(c_S_A*c_S_B)**2)
    f_AB = c_S_A*(1+8*(c_S_A*c_S_B)**2)
    # Calculate new volumes using surface fractions (which use original
    # volumes).
    vol_A = vol0_A*(1+a_A*f_BA*(phi[0]-phi[1]))
    vol_B = vol0_B*(1+a_B*f_AB*(phi[1]-phi[0]))
    # Recalculate surface concentrations using new volumes.
    c_S_A = (1-xB)*vol_A/((1-xB)*vol_A+xB*vol_B)
    c_S_B = xB*vol_B/((1-xB)*vol_A+xB*vol_B)
    # Recalculate surface fractions for ordered compounds using new volumes.
    f_BA = c_S_B*(1+8*(c_S_A*c_S_B)**2)
    f_AB = c_S_A*(1+8*(c_S_A*c_S_B)**2)
    D_htrans = xB*htrans[1]+(1-xB)*htrans[0]
    H_ord = (Gamma*(1-xB)*xB*vol_A*vol_B*(1+8*(c_S_A*c_S_B)**2)/
            ((1-xB)*vol_A+xB*vol_B) + D_htrans)
    return round(H_ord*0.01036427, 2)
    #return int(round(H_ord))

def get_energy(composition):
    '''
    composition must be a dictionary of elements and amounts. If not binary, or
    if any element isn't parameterized, returns None.

    composition = {'A':Nx, 'B':Ny, 'C':Nz}
    '''
    # validate composition
    if len(composition) != 2:
        return None
    filename = "/home/sjk648/bin/qmpy/utils/miedema.dat"
    elements = input_elements(filename)

    if not all( elements[k] for k in composition ):
        return None
    natoms=sum(composition.values())
    A,B = composition.keys()
    xB = float(composition[B])/natoms
    return H_form_ord(A,B,xB,elements)



from collections import defaultdict
import itertools
import numpy as np
import re
from qmpy.element import *
from my_math import *

'''
A 'formula' or 'name' is FeO2, Ni3B, etc...
A 'composition' or 'comp' is {'Fe':1, 'O':2}, {'Ni":3, 'B':1}, etc...
'''

sfind = re.compile('[A-Z][a-wyz]?[0-9\.x]*')
#TODO
#   - fix parse_formula

def stupid_order(elt):
    if elt == 'O':
        return 'zzzzzz'
    elif elt in [ 'F', 'Br', 'Cl', 'I' ]:
        return 'zzzzz'
    elif elt == 'H':
        return 'zzzz'
    elif elt in [ 'P', 'S', 'N', 'C', 'B', 'Si' ]:
        return 'zzz'+elt
    elif elt in [ 'Li', 'Na', 'K', 'Rb', 'Cs' ]:
        return 'AAA'+elt
    elif elt in ['Be', 'Mg', 'Ca', 'Sr', 'Ba']:
        return 'AA'+elt
    else:
        return elt

def comp_to_latex(comp, special='reduce'):
    if special=='reduce':
        divisor = get_gcd(comp.values())
        comp = dict((k,v/divisor) for k,v in comp.items())
    name = ""
    for key in sorted(comp.keys(), key=lambda x: stupid_order(x)):
        if comp[key] == 1:
            n = ""
        else:
            if comp[key]:
                n = '_{%s}' % int(comp[key])
        name += key+n
    if name == 'LiHO':
        name = 'LiOH'
    return name

def comp_to_name(comp, special=None):
    if special=='reduce':
        divisor = get_gcd(comp.values())
        comp = dict((k,v/divisor) for k,v in comp.items())
    name = ""
    for key in sorted(comp.keys()):
        if comp[key] == 1:
            n = ""
        else:
            n = str(int(comp[key]))
        name += key+n
    if name == 'LiHO':
        name = 'LiOH'
    return name

def comp_to_formula(comp, special=None):
    if special=='reduce':
        divisor = get_gcd(comp.values())
        comp = dict((k,v/divisor) for k,v in comp.items())
    name = ""
    for key in sorted(comp.keys()):
        name += '%s%s,' % (key, int(comp[key]))
    name = name.strip(',')
    return name

def name_to_comp(name, special=None):
    comp = defaultdict(float)
    for elt in sfind.findall(name):
        e = ''
        n = ''
        for char in elt:
            if char.isalpha():
                if char == 'x':
                    n = 0
                else:
                    e += char
            else:
                n += char
        if n == '': 
            n = 1
        if e == 'D' or e == 'T':
            e = 'H'
        comp[e] = float(n)
    divisor = 1.
    if special == 'reduce':
        divisor = get_gcd(comp.values())
    if special == 'unit':
        divisor = sum(comp.values())
    return defaultdict(float, dict((k,v/divisor) for k, v in comp.items()))

def parse_formula(formula):
    formulae = []
    sfind = re.compile('({[^}]*}[:0-9x\.]*|[A-Z][a-wyz]?[:0-9x\.]*)')
    matches = sfind.findall(formula)
    for term in matches:
        if '{' in term:
            symbols, amt = term.replace('{','').split('}')
            symbols = symbols.split(':')
            elts = []
            for symbol in symbols:
                if ElementSet.objects.filter(name=symbol).exists():
                    eset = ElementSet.objects.get(name=symbol)
                    elts += [ e.symbol for e in eset.elements.all()]
                else:
                    elts += [ symbol ]
            formulae.append([ e+amt for e in elts ])
        else:
            formulae.append([term])
    return [ name_to_comp(''.join(ref)) for ref in itertools.product(*formulae) ]



from django.db import models
import json

from utils import *
from element import Element, ElementSet
from thermodynamics import Formation
from thermopy import PhaseSpace

##TODO
## - define properties seperate from methods

class Composition(models.Model):
    '''
    A composition object represents a point in composition space - i.e. it has
    a defined elemental composition, but not a defined structure. As such it
    can correspond to many structures, many experiments, and has a single
    valued meidema energy and hull energy.
    '''

    formula = models.CharField(primary_key=True, max_length=255, db_index=True)
    name = models.CharField(max_length=255, db_index=True)
    keywords = models.CharField(max_length=255, null=False)

    element = models.ManyToManyField('Element', null=True)
    ntypes = models.IntegerField(null=True)

    ### other stuff

    mass = models.FloatField(blank=True, null=True)

    ### thermodyanamic stuff

    meidema = models.FloatField(blank=True, null=True)
    matproj = models.FloatField(blank=True, null=True)
    oqmd = models.FloatField(blank=True, null=True)
    icsd = models.FloatField(blank=True, null=True)
    expt = models.FloatField(blank=True, null=True)

    class Meta:
        db_table = 'compositions'

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.name = comp_to_name(self.comp, special='reduce')
        self.ntypes = len(self.comp)
        self.meidema = get_energy(self.comp)
        super(Composition, self).save(*args, **kwargs)
        if not self.element.count():
            for elt in self.comp:
                self.element.add(Element.get(elt))

    @property
    def summary(self):
        summary = 'Summary of %s:\n' % self.name
        sums = {}
        for k, v in self.distinct.items():
            s = '   - %s (SG# %s, %s atoms): ' % (v.path, k[0], k[1])
            if not v.tasks:
                print 'Not queued for some reason'
            elif v.holds:
                s += 'held for %s' % ', '.join(v.holds.split())
                sums[s] = 30000
            elif v.tasks[0].state==1:
                s += 'running'
                sums[s] = 20000
            elif not v.oqmd:
                sums[s] = 10000
            else:
                s += '%s' % v.oqmd.energy_pa
                sums[s] = v.oqmd.energy_pa

        sums = sorted( sums.keys(), key=lambda x: sums[x])
        summary += '\n'.join(sums)
        return summary

    @staticmethod
    def get(composition):
        if isinstance(composition, basestring):
            composition = name_to_comp(composition)
        cs, created = Composition.objects.get_or_create(
                formula=comp_to_formula(composition, special='reduce'))
        return cs

    @staticmethod
    def get_space(bounds):
        if isinstance(bounds, list):
            bounds = '-'.join(bounds)
        space = PhaseSpace(bounds, load=None)
        phases = []
        comps = Composition.objects.none()
        elts = [ Element.get(elt) for elt in space.space ]
        for i in range(len(space.space)):
            for combi in itertools.combinations(elts, r=i+1):
                for elt in combi:
                    comps |= elt.composition_set.filter(ntypes=i+1)
        comps = comps.values_list('formula', flat=True)
        for comp in comps:
            cdict = name_to_comp(comp)
            if space.in_space(cdict):
                phases.append(Composition.objects.get(formula=comp))
        return phases

    @property
    def calculable(self):
        return list(self.entry_set.filter(holds=''))

    @property
    def distinct(self):
        if not hasattr(self, '_distinct'):
            self.get_distinct(icsd=False)
        return self._unique

    @property
    def icsd_distinct(self):
        if not hasattr(self, '_distinct'):
            self.get_distinct(icsd=True)
        return self._unique

    @property
    def duplicates(self):
        if not hasattr(self, '_duplicates'):
            self.get_distinct(icsd=True)
        return self._duplicates

    def get_distinct(self, icsd=False, calculable=True, high_pressure=True):
        unique = {}
        duplicates = defaultdict(list)
        entries = self.entry_set.filter(structure__label='source')
        if not high_pressure:
            entries = entries.exclude(keywords__contains='pressure')
        if icsd:
            entries = entries.filter(path__contains='icsd')
        if calculable:
            entries = entries.filter(holds='')
        for entry in entries:
            inp = entry.input
            if not inp.spacegroup:
                inp.find_symmetry()
                inp.save()
            if icsd:
                sg = entry.icsd.spacegroup
            else:
                sg = inp.spacegroup

            duplicates[(sg, inp.natoms)].append(entry)
            if not (sg, inp.natoms) in unique:
                unique[(sg, inp.natoms)] = entry
            if (entry.is_done('standard') and not
                    unique[(sg, inp.natoms)].is_done('standard')):
                unique[(sg, inp.natoms)] = entry

        for entry in unique.values():
            entry.unique = True
            entry.save()

        self._unique = unique
        self._duplicates = duplicates

    @property
    def comp(self):
        return defaultdict(float, 
                name_to_comp(self.formula))

    @property
    def nom_comp(self):
        return defaultdict(float, 
                name_to_comp(self.formula, special='reduce'))

    @property
    def unit_comp(self):
        return defaultdict(float,
                name_to_comp(self.formula, special='unit'))

    def set_thermo(self, fit='standard'):
        expt = self.experiment_set.filter(dft=True).order_by('delta_e')
        if expt.exists():
            self.matproj = expt[0].delta_e

        expt = self.experiment_set.filter(dft=False).order_by('delta_e')
        if expt.exists():
            self.expt = expt[0].delta_e
        if expt.filter(source='ssub').exists():
            self.expt = expt.filter(source='ssub')[0].delta_e
        if expt.filter(comment='correction').exists():
            self.expt = expt.filter(comment='correction')[0].delta_e

        forms = self.formation_set.filter(fit=fit).order_by('delta_e')
        if forms.exists():
            self.oqmd = forms[0].delta_e

        f1 = forms.filter(calculation__path__contains='icsd')
        f2 = f1.order_by('delta_e')
        if f2.exists():
            self.icsd = f2[0].delta_e

        self.meidema = get_energy(self.comp)

    @property
    def gs(self):
        calcs = self.calculation_set.filter(settings='standard', done=True)
        calcs = calcs.filter(path__contains='icsd')
        calcs = calcs.order_by('energy_pa')
        if calcs.exists():
            return calcs[0]
        else:
            return None

    @property
    def hubbard(self):
        return any( hub for hub in self.gs.hubbards )

    @property
    def volume(self):
        if self.gs is not None:
            return self.gs.volume_pa
        else:
            return None

    ## keywords
    def add_keyword(self, keyword):
        keywords = self.get_keywords()
        if keyword not in keywords:
            keywords.append(keyword)
        self.keywords = ' '.join(keywords)

    def get_keywords(self):
        return self.keywords.strip().split(' ')

    def remove_keyword(self, keyword):
        keywords = self.keywords
        keywords.remove(keyword)
        self.keywords = ' '.join(keywords)

    @property
    def latex(self):
        return comp_to_latex(self.comp)



'''Calculation instances correspond to a particular VASP calculation and
contains:
    -The necessary inputs to recreate the calculation
    -The interesting outputs
    -Make needed changes to fix problems encountered while running

'''

import os
import copy
import json
import time
import numpy as np
from os.path import exists, isfile, isdir

from django.db import models

from dev_qmpy.models.composition import Composition
from dev_qmpy.models.element import *
from dev_qmpy.models.structure import Structure, Atom
from dev_qmpy.utils import *

class VaspError(Exception):
    '''General problem with vasp calculation.'''

class VaspZPOTRFError(VaspError):
    '''VASP encountered ZPOTRF error'''

class VaspSGRCONError(VaspError):
    '''VASP encountered SGRCON error'''

class VaspINVGRPError(VaspError):
    '''VASP encountered INVGRP error'''

class VaspBRIONSError(VaspError):
    '''VASP encountered a BRIONS problems error'''

class VaspFEXCFError(VaspError):
    '''VASP encountered'''

class VaspFEXCPError(VaspError):
    '''VASP encountered'''

class VaspPRICELError(VaspError):
    '''VASP encountered'''

class VaspEDDDAVError(VaspError):
    '''VASP encountered'''

class VaspNotHermitianError(VaspError):
    '''VASP encountered'''

class VaspBRMIXError(VaspError):
    '''VASP encountered'''

class VaspIncompleteError(VaspError):
    '''VASP did not complete the calculation'''

class VaspElectronicConvergenceError(VaspError):
    '''VASP did not reach electronic convergence'''

class VaspIonicConvergenceError(VaspError):
    '''VASP did not reach ionic convergence'''

class VaspDOSCARError(VaspError):
    '''Not able to read the DOSCAR correctly'''

class VaspInputError(VaspError):
    '''Not able to read inputs correctly'''

class VaspAttemptsError(VaspError):
    '''VASP failed to successfully complete too many times in a row'''

class VaspIonsCloseError(VaspError):
    '''VASP found ions too close together'''

class VaspMultirunError(VaspError):
    '''VASP was run multiple times over simultaneously, leading to unparseable
    outputs'''

class VaspBandsError(VaspError):
    '''VASP encountered fill bands at some k-points, at some ionic step'''

class Calculation(models.Model):
    # labeling
    settings = models.CharField(db_index=True, max_length=15, 
            null=True, blank=True)
    hubbard_set = models.ManyToManyField('Hubbard')
    potential_set = models.ManyToManyField('Potential')

    entry = models.ForeignKey('Entry', db_index=True, 
            null=True, blank=True)
    path = models.CharField(max_length=255, null=True, db_index=True)
    args = models.CharField(max_length=255, default=json.dumps([]))
    attempt = models.IntegerField(default=0)
    done = models.BooleanField(blank=True)

    composition = models.ForeignKey('Composition', 
            null=True, blank=True)
    name = models.CharField(max_length=255, db_index=True, 
            null=True, blank=True)
    formula = models.CharField(max_length=255, db_index=True, 
            null=True, blank=True)

    # inputs
    #   -Structure
    input = models.ForeignKey(Structure, related_name='calculations',
            null=True, blank=True)

    #   -General
    prec = models.CharField(max_length=10, null=True, blank=True)
    istart = models.IntegerField(default=1)
    icharg = models.IntegerField(default=1)
    ispin = models.IntegerField(default=1)
    lsorbit = models.BooleanField(default=False)
    nelect = models.FloatField(blank=True, null=True)
    ldau = models.BooleanField(default=False)

    # parallization
    npar = models.IntegerField(default=2)
    ncore = models.IntegerField(blank=True, null=True)
    nsim = models.IntegerField(blank=True, null=True)
    lplane = models.IntegerField(default=True)
    kpar = models.IntegerField(blank=True, null=True)

    #   -electronic relaxation 1
    encut = models.FloatField(null=True, blank=True)
    enaug = models.FloatField(null=True, blank=True)
    nelm = models.IntegerField(default=60)
    nelmin = models.IntegerField(default=2)
    ediff = models.FloatField(default=0.1E-3)
    lreal = models.CharField(max_length=10, default='false')
    lmaxmix = models.IntegerField(default=2)
    voskown = models.IntegerField(default=0)

    #   -Ionic relaxation
    ediffg = models.FloatField(default=0.1E-02)
    nsw = models.IntegerField(default=0)
    ibrion = models.IntegerField(default=0)
    isif = models.IntegerField(default=2)
    isym = models.IntegerField(default=1)
    pstress = models.FloatField(default=0.0)
    potim = models.FloatField(default=0.5)
    symprec = models.FloatField(default=1e-5)

    #   -DOS related
    nbands = models.IntegerField(null=True, blank=True)
    ismear = models.IntegerField(default=0)
    sigma = models.FloatField(default=0.2)

    #   -electronic relaxation 2
    algo = models.CharField(max_length=15, default='normal')
    amix = models.FloatField(default=0.4)
    bmix = models.FloatField(default=1.0)
    maxmix = models.IntegerField(default=-45)

    #   -write fields
    lwave = models.BooleanField(default=False)
    lcharg = models.BooleanField(default=True)
    lvtot = models.BooleanField(default=False)
    lvhar = models.BooleanField(default=True)
    lelf = models.BooleanField(default=True)
    lorbit = models.IntegerField(default=11)

    #   -dipol fields
    ldipol = models.BooleanField(default=False)
    idipol = models.IntegerField(default=0)
    epsilon = models.FloatField(null=True, blank=True)

    #   - custom fields
    custom = models.TextField(blank=True, null=True)
    
    #kpoints
    kppra = models.IntegerField(null=True, blank=True)
    irreducible_kpoints = models.IntegerField(null=True, blank=True)
    gamma = models.BooleanField(default=True)
    
    # outputs
    energy = models.FloatField(null=True, blank=True)
    energy_pa = models.FloatField(null=True, blank=True)
    magmom = models.FloatField(default=0)
    magmom_pa = models.FloatField(default=0)
    band_gap = models.FloatField(blank=True, null=True)
    fermi = models.FloatField(blank=True, null=True)
    volume = models.FloatField(null=True, blank=True)
    volume_pa = models.FloatField(null=True, blank=True)
    output = models.ForeignKey(Structure, 
            related_name='origin',
            null=True, blank=True)

    niter = models.IntegerField(default=0)
    econverged = models.NullBooleanField(null=True)
    fconverged = models.NullBooleanField(null=True)
    voluntary = models.NullBooleanField(null=True)
    runtime = models.FloatField(default=0)
    errors = models.CharField(max_length=255, default=json.dumps([]))

    class Meta:
        db_table = 'calculations'

    def __str__(self):
        if self.settings is not None:
            return '%s : %s' % (comp_to_name(self.input.comp), self.settings )
        else:
            return comp_to_name(self.input.comp)

    def __eq__(self, other):
        diff = []
        for attr in ['encut', 'prec', 'lreal', 'potentials']:
            if getattr(self, attr) is None:
                continue
            elif getattr(other, attr) is None:
                diff.append([ attr, 
                    getattr(self, attr),
                    getattr(other, attr)])
            elif getattr(self, attr) != getattr(other, attr):
                diff.append([ attr, 
                    getattr(self, attr),
                    getattr(other, attr)])

        if other.ispin < self.ispin:
            diff.append(['ispin', self.ispin, other.ispin])

        if self.hubbard:
            if not other.hubbard:
                diff.append(['hubbard', self.hubbards, other.hubbards ])
            else:
                for hub in self.hubbards:
                    if hub and hub not in other.hubbards:
                        diff.append(['hubbard', hub, None])
                for hub in other.hubbards:
                    if hub and hub not in self.hubbards:
                        diff.append(['hubbard', None, hub])
        if self.nsw > 0:
            if other.nsw <= 1:
                diff.append(['nsw', self.nsw, other.nsw])
        self.diff = diff
        other.diff = diff
        if diff:
            return False
        else:
            return True


    ### magic functions and other "builtins"

    def save(self, *args, **kwargs):
        if self.output is not None:
            self.output.save()
            self.output = self.output
            self.output.entry = self.entry
        if self.input is not None:
            self.input.save()
            self.input = self.input
            self.input.entry = self.entry
            self.composition = self.input.composition
        self.name = comp_to_name(self.input.comp, special='reduce')
        super(Calculation, self).save(*args, **kwargs)
        if hasattr(self, 'new_hubbards'):
            for hub in self.new_hubbards:
                self.hubbard_set.add(hub)
            del self.new_hubbards
        if hasattr(self, 'new_potentials'):
            for pot in self.new_potentials:
                self.potential_set.add(pot)
            del self.new_potentials

    def add_potential(self, potential):
        if not hasattr(self, 'new_potentials'):
            self.new_potentials = []
        self.new_potentials.append(potential)

    def add_hubbard(self, hubbard):
        if not hasattr(self, 'new_hubbards'):
            self.new_hubbards = []
        self.new_hubbards.append(hubbard)

    def add_potential(self, pot):
        if not hasattr(self, 'new_potentials'):
            self.new_potentials = []
        self.new_potentials.append(pot)

    @property
    def potentials(self):
        if hasattr(self, 'new_potentials'):
            return sorted(set(self.new_potentials), key=lambda x: x.element.symbol)
        elif self.id is None:
            return []
        return sorted(self.potential_set.distinct().all(), key=lambda x: x.element.symbol)

    @property
    def hubbards(self):
        if hasattr(self, 'new_hubbards'):
            return sorted(set(self.new_hubbards), key=lambda x:x.element.symbol)
        elif self.id is None:
            return []
        return sorted(self.hubbard_set.all(), key=lambda x:x.element.symbol)

    def get_args(self):
        return json.loads(self.args)

    def set_args(self, args):
        self.args = json.dumps(args)

    @property
    def comp(self):
        return defaultdict(int,
                name_to_comp(self.formula))

    @property
    def unit_comp(self):
        return defaultdict(float,
                name_to_comp(self.formula, special='unit'))

    @property
    def hubbard(self):
        return any( h for h in self.hubbards )

    ## setup stuff 
    @property
    def INCAR(self):
        incar = '# General\n'
        if not self.prec is None: 
            incar += ' PREC = %s\n' % self.prec.upper()
        if not self.istart is None:
            incar += ' ISTART = %s\n' % self.istart
        if not self.icharg is None:
            incar += ' ICHARG = %s\n' % self.icharg
        incar += self.input.INCAR
        if self.lsorbit:
            incar += ' LSORBIT = .TRUE.\n'
        if self.nelect is not None:
            incar += ' NELECT = %s\n' % self.nelect

        incar += '\n# parallization\n'
        incar += ' NPAR = %s\n' % self.npar
        if self.lplane:
            incar += ' LPLANE = .TRUE.\n'
        else:
            incar += ' LPLANE = .FALSE.\n'
        if self.nsim:
            incar += ' NSIM = %s\n' % self.nsim
        if self.ncore:
            incar += ' NCORE = %s\n' % self.ncore
        if self.kpar:
            incar += ' KPAR = %s\n' % self.kpar

        incar += '\n# electronic relaxation 1\n'
        if self.encut:
            incar += ' ENCUT = %s\n' % self.encut
        #if self.enaug:
        #    incar += ' ENAUG = %s\n' % self.enaug
        incar += ' NELM = %s\n' % self.nelm
        incar += ' NELMIN = %s\n' % self.nelmin
        incar += ' EDIFF = %s\n' % self.ediff
        if self.lreal is not None:
            incar += ' LREAL = %s\n' % self.lreal.upper()
        if self.voskown: incar += ' VOSKOWN = TRUE\n'

        incar += '\n# Ionic relaxation\n'
        incar += ' IBRION = %s\n' %  self.ibrion
        incar += ' ISIF = %s\n' % self.isif
        incar += ' NSW = %s\n' % self.nsw
        incar += ' ISYM = %s\n' % self.isym
        if not self.symprec is None:
            incar += ' SYMPREC = %s\n' % self.symprec
        incar += ' POTIM = %s\n' % self.potim

        incar += '\n# DOS related\n'
        if not self.nbands is None: incar += ' NBANDS = %s\n' % self.nbands
        if not self.ismear is None: incar += ' ISMEAR = %s\n' % self.ismear
        if not self.sigma is None: incar += ' SIGMA = %s\n' % self.sigma

        incar += '\n# electronic relaxation 2\n'
        incar += ' ALGO = %s\n' % self.algo.upper()
        incar += ' AMIX = %s\n' % self.amix
        incar += ' BMIX = %s\n' % self.bmix
        incar += ' MAXMIX = %s\n' % self.maxmix

        incar += '\n# write fields\n'
        if self.lwave: 
            incar += ' LWAVE = .TRUE.\n'
        else:
            incar += ' LWAVE = .FALSE.\n'
        if self.lcharg: 
            incar += ' LCHARG = .TRUE.\n'
        else:
            incar += ' LCHARG = .FALSE.\n'
        if self.lvtot: 
            incar += ' LVTOT = .TRUE.\n'
        else:
            incar += ' LVTOT = .FALSE\n'
        if self.lvhar:
            incar += ' LVHAR = .TRUE.\n'
        else:
            incar += ' LVHAR = .FALSE.\n'
        if self.lelf: 
            incar += ' LELF = .TRUE.\n'
        else:
            incar += ' LELF = .FALSE.\n'
        incar += ' LORBIT = %s\n' % int(self.lorbit)

        if self.ldipol:
            incar += '\n# dipole fields\n'
            incar += ' LDIPOL = .TRUE.\n'
            incar += ' IDIPOL = %s\n' % self.idipol
            incar += ' EPSILON = %s\n' % self.epsilon


        if not self.custom is None:
            incar += '\n# custom\n'   
            incar += '%s\n' % self.custom.upper()

        if  any(hub for hub in self.hubbards):
            incar += '\n# LDA+U fields\n'
            incar += ' LDAU = .TRUE.\n'
            incar += ' LDAUPRINT = 1\n'
            uvals = ' '.join(str(hub.ldau_u) for hub in self.hubbards)
            jvals = ' '.join('0' for hub in self.hubbards)
            lvals = ' '.join(str(hub.ldau_l) for hub in self.hubbards)
            incar += ' LDAUU = %s\n' % uvals
            incar += ' LDAUJ = %s\n' % jvals
            incar += ' LDAUL = %s\n' % lvals
        return incar

    @INCAR.setter
    def INCAR(self, incar):
        custom = ''
        magmoms = []
        ldauus = []
        ldauls = []
        ldaujs = []
        for line in open(incar):
            line = line.lower()
            # general
            if 'prec' in line:
                self.prec = line.split()[-1]
            elif 'encut' in line:
                self.encut = line.split()[-1]
            elif 'isif' in line:
                self.isif = line.split()[-1]
            elif 'ibrion' in line:
                self.ibrion = line.split()[-1]
            elif 'nbands' in line:
                self.nbands = line.split()[-1]
            elif 'lreal' in line:
                self.lreal = line.split()[-1]
            elif 'ismear' in line:
                self.ismear = line.split()[-1]
            elif 'sigma' in line:
                self.sigma = line.split()[-1]
            elif 'ldauu' in line:
                ldauus = line.split()[2:]
            elif 'ldaul' in line:
                ldauls = line.split()[2:]
            elif 'magmom' in line:
                magmoms = line.split()[2:]
            else:
                custom += line

        if self.input is not None:
            atom_types = []
            for atom in self.input.atoms:
                if atom.element.symbol in atom_types:
                    continue
                atom_types.append(atom.element.symbol)
            if ldauus and ldauls:
                #assert len(ldauus) == len(ldauls) == len(atom_types)
                for elt, u, l in zip(atom_types, ldauus, ldauls):
                    hub, created = Hubbard.objects.get_or_create(element_id=elt,
                            ldau_u=u, ldau_l=l, with_elt=None)
                    #if created:
                    #    hub.save()
                    self.add_hubbard(hub)
            if magmoms:
                real_moms = []
                for mom in magmoms:
                    if '*' in mom:
                        num, mom = mom.split('*')
                        real_moms += [mom]*int(num)
                    else:
                        real_moms.append(mom)
                for atom, mom in zip(self.input.atoms, real_moms):
                    atom.magmom = float(mom)
                    if atom.id is not None:
                        atom.save()

    @property
    def KPOINTS(self):
        n = float(len(self.input.atoms))
        refr = self.input.relative_rec_lat
        scale = 1.0
        kpts = np.ones(3)
        while n*np.product(kpts) < self.kppra:
            prev_kpts = kpts.copy()
            refk = np.array(np.ones(3)*refr)*scale
            kpts = np.array(map(np.ceil, refk))
            scale += 1
        if self.kppra - np.product(prev_kpts)*n < np.product(kpts)*n - self.kppra:
            kpts = prev_kpts.copy()
        if self.gamma:
            kpoints = 'KPOINTS \n0 \nGamma\n'
        else:
            kpoints = 'KPOINTS \n0 \nMonkhost-Pack\n'
        kpoints += ' '.join( str(int(k)) for k in kpts ) + '\n'
        kpoints += '0 0 0'
        return kpoints

    @KPOINTS.setter
    def KPOINTS(self, kpoints):
        raise NotImplementedError

    @property
    def POTCAR(self):
        potstr = ''
        for pot in self.potentials:
            potstr += pot.potcar
            potstr += ' End of Dataset\n'
        return potstr

    @POTCAR.setter
    def POTCAR(self, potcar):
        pots = Potential.read_potcar(potcar)
        for pot in pots:
            self.add_potential(pot)

    @property
    def POSCAR(self):
        return self.input.POSCAR

    @POSCAR.setter
    def POSCAR(self, poscar):
        self.input = Structure.create(poscar)

    def clear_outputs(self):
        if not os.path.exists(self.path):
            return
        for file in os.listdir(self.path):
            if os.path.isdir(self.path+'/'+file):
                continue
            if file in ['INCAR', 'POSCAR', 'KPOINTS', 'POTCAR',
                    'CHGCAR.gz']:
                continue
            os.unlink('%s/%s' % (self.path, file))

    def unpack(self):
        if 'OUTCAR.gz' in os.listdir(self.path):
            os.system('gunzip -f %s/OUTCAR.gz' % self.path)

    def pack(self):
        for file in os.listdir(self.path):
            if file in ['OUTCAR', 'CHGCAR', 'CHG', 'PROCAR', 'LOCPOT', 'ELFCAR']:
                os.system('gzip -f %s' % self.path+'/'+file)

    def copy(self):
        calc = copy.deepcopy(self)
        calc.id = None
        if self.input is not None:
            calc.input = self.input.copy()
        if self.output is not None:
            calc.output = self.output.copy()
        for p in self.potentials:
            calc.add_potential(p)
        for h in self.hubbards:
            calc.add_hubbard(h)
        return calc

    @staticmethod
    def read(path):
        '''
        Reads the outcar specified by the objects path. Populates input field
        values, as well as outputs, in addition to finding errors and
        confirming convergence.

        Example:
        >>> calc = Calculation.read('/path/to/calculation')
        '''
        if Calculation.objects.filter(path=path, done=True).exists():
            return Calculation.objects.get(path=path, done=True)

        calc = Calculation()
        calc.path = path
        try:
            calc.read_outcar()
            calc.check_stdout()
        except Exception, err:
            calc.add_error('exception: %s' % err)
            print '   -Exception:', err
        return calc

    def read_outcar(self):
        self.unpack()
        if not exists(self.path):
            return
        if not exists(self.path+'/OUTCAR'):
            return
        if len(open(self.path+'/OUTCAR').readlines()) < 100:
            return
        if not exists(self.path+'/CONTCAR'):
            return

        prev_attempts = []
        for dir in [d for d in os.listdir(self.path) if isdir(self.path+'/'+d)]:
            prev_attempts.append(d)
        self.attempt = len(prev_attempts)+1
        if self.attempt > 3:
            self.add_error('attempts')
        elif self.attempt > 30:
            self.add_error('3xattempts')

        outcar = open('%s/OUTCAR' % self.path,'r').readlines()
        stresses = []
        runcount = 0
        input_cell = []
        input_coords = []
        cell = []
        species = []
        coords = []
        forces = []
        magmoms = []
        charges = []
        look_for_total = False
        esteps = 0
        this_pot = {}
        ldauls = []
        ldauus = []
        atom_types = []
        for n, line in enumerate(outcar):
            line = line.strip()
            ### error lines
            if 'between ions is very small' in line:
                self.add_error('ions_close')
            elif 'Voluntary' in line:
                self.voluntary = True
                runcount += 1
                if runcount > 1:
                    self.add_error('multiruns')
            elif 'reached required accuracy' in line:
                self.fconverged = True
            elif '%s(' % self.nsw in line:
                self.add_error('ionic_convergence')
            elif 'highest band is occupied' in line:
                self.add_error('bands')

            ### input structure lines
            elif 'ions per type' in line:
                atom_types = []
                counts = line.split()[4:]
                int_counts = [ int(count) for count in counts ]
                assert len(int_counts) == len(species)
                for n, e in zip(int_counts, species):
                    atom_types += [Element.get(e)]*n
            elif 'nearest neighbor table' in line:
                i = 1
                cdata = outcar[n+i]
                while not 'LATTYP' in cdata:
                    if '-' in cdata:
                        input_coords.append([ float(v) for v in 
                            cdata.split('-')[0].split()[1:] ])
                    i += 1
                    cdata = outcar[n+i]
            elif 'A1 = (' in line:
                vec = line.split('(')[1].strip(')')
                vals = [ float(v.strip()) for v in vec.split(',') ]
                input_cell.append(vals)
            elif 'A2 = (' in line:
                vec = line.split('(')[1].strip(')')
                vals = [ float(v.strip()) for v in vec.split(',') ]
                input_cell.append(vals)
            elif 'A3 = (' in line:
                vec = line.split('(')[1].strip(')')
                vals = [ float(v.strip()) for v in vec.split(',') ]
                input_cell.append(vals)

            ### general options
            elif 'PREC' in line:
                self.prec = line.split()[2]
            elif 'ENCUT' in line:
                self.encut = line.split()[2]
            elif 'ENAUG' in line:
                self.enaug = line.split()[2]
            elif 'ISTART' in line:
                self.istart = line.split()[2]
            elif 'ISPIN' in line:
                self.ispin = int(line.split()[2])

            # electronic relaxation 1
            elif 'NELM' in line:
                self.nelm = int(line.split()[2].rstrip(';'))
                self.nelmin = int(line.split()[4].rstrip(';'))
            elif 'LREAL  =' in line:
                lreal = line.split()[2]
                if lreal == 'F':
                    self.lreal = 'false'
                elif lreal == 'A':
                    self.lreal = 'auto'
                elif lreal == 'T':
                    self.lreal = 'auto'

            # ionic relaxation 
            elif 'EDIFF  =' in line:
                self.ediff = line.split()[2]
            elif 'ISIF' in line:
                self.isif = line.split()[2]
            elif 'IBRION' in line:
                self.ibrion = line.split()[2]
            elif 'NSW' in line:
                self.nsw = line.split()[2].rstrip(';')
                if self.nsw <= 1:
                    self.fconverged = True
            elif 'PSTRESS' in line:
                self.pstress = line.split()[1]
            elif 'POTIM' in line:
                self.potim = line.split()[2]

            # DOS Flags
            elif 'ISMEAR' in line:
                line = line.split()
                self.ismear = line[2].rstrip(';')
                self.sigma = line[5]
            elif 'NBANDS=' in line:
                if not 'INCAR' in line:
                    self.nbands = int(line.split()[-1])

            # write flags
            elif 'LCHARG' in line:
                if line.split()[2] == 'F':
                    self.lcharg = False
            elif 'LWAVE' in line:
                if line.split()[2] == 'F':
                    self.lwave = False
            elif 'LVTOT' in line:
                if line.split()[2] == 'T':
                    self.lvtot = True
            elif 'LELF' in line:
                if line.split()[2] == 'T':
                    self.lelf = True
            elif 'LORBIT' in line:
                self.lorbit = line.split()[2]

            # electronic relaxation 2
            elif 'AMIX     =' in line:
                self.amix = line.split()[2].rstrip(';')
                self.bmix = line.split()[5]
            elif 'ALGO' in line:
                val = int(line.split()[2])
                if val == 38:
                    self.algo = 'normal'
                elif val == 68:
                    self.algo = 'fast'
                elif val == 48:
                    self.algo = 'very_fast'
                elif val == 58:
                    self.algo = 'all'
                elif val == 53:
                    self.algo = 'default'
            elif 'MAXMIX=' in line:
                self.maxmix = line.split()[-1]

            # dipole flags
            elif 'LDIPOL' in line:
                if line.split()[2] == 'T':
                    self.ldipol = True
            elif 'IDIPOL' in line:
                self.idipol = line.split()[2]
            elif ' EPSILON=' in line:
                self.epsilon = line.split()[1]

            # potentials
            elif 'POTCAR:' in line:
                this_pot = {'name':line.split()[2]}
                if not 'VRHFIN' in outcar[n+1]:
                    elt = line.split()[2].split('_')[0]
                    species.append(elt)
            elif 'Description' in line:
                pot = Potential.objects.get(**this_pot)
                self.add_potential(pot)
            elif 'LEXCH' in line:
                key = line.split()[2]
                if key == '91':
                    this_pot['xc'] = 'GGA'
                elif key == 'CA':
                    this_pot['xc'] = 'LDA'
                elif key == 'PE':
                    this_pot['xc'] = 'PBE'
            elif 'LULTRA' in line:
                key = line.split()[2]
                this_pot['us'] = ( key == 'T' )
            elif 'LPAW' in line:
                key = line.split()[2]
                this_pot['paw'] = ( key == 'T' )
            # hubbards
            elif 'LDAUL' in line:
                self.ldau = True
                ldauls = [ int(v) for v in line.split()[7:] ]
            elif 'LDAUU' in line:
                ldauus = [ float(v) for v in line.split()[7:] ]

            ### output structure lines
            if 'direct lattice vectors' in line:
                cell = []
                for i in range(3):
                    vec = []
                    vec.append(float(outcar[n+1+i][0:16]))
                    vec.append(float(outcar[n+1+i][16:29]))
                    vec.append(float(outcar[n+1+i][29:44]))
                    cell.append(vec)
            if self.ispin == 2 and 'magnetization (x)' in line:
                magmoms = []
                for ind in range(len(atom_types)):
                    magmoms.append(ffloat(outcar[n+4+ind].split()[-1]))
            if 'total charge' in line and not 'along one line' in line:
                charges = []
                for ind in range(len(atom_types)):
                    charges.append(ffloat(outcar[n+4+ind].split()[-1]))
            if 'POSITION' in line:
                coords = []
                forces = []
                for ind in range(len(atom_types)):
                    l = outcar[n+ind+2]
                    try:
                        l = l.split()
                    except:
                        l = [l[0:12], l[13:25], l[26:38], 
                                l[39:55], l[56:69],l[70:84]]
                    coords.append([ float(v) for v in l[0:3] ])
                    forces.append([ float(v) for v in l[3:6] ])

            ### output property lines
            if 'free  energy' in line:
                self.energy = ffloat(line.split()[4])
                self.energy_pa = self.energy/len(atom_types)
            if self.ispin == 2 and 'number of electron' in line:
                if len(line.split()) == 6:
                    self.magmom = abs(ffloat(line.split()[-1]))
                    self.magmom_pa = self.magmom/len(atom_types)
            if 'LOOP+' in line:
                if not len(line.split()) == 7:
                    continue
                self.runtime += ffloat(line.split()[-1].lstrip('time'))
            if 'irreducible' in line:
                self.irreducible_kpoints = int(line.split()[1])
            if 'volume of cell' in line:
                self.volume = ffloat(line.split()[-1])
                self.volume_pa = self.volume/len(atom_types)
            if 'Iteration' in line:
                if len(line.split()) == 5:
                    l = line.split()
                    self.niter = int(l[2].rstrip('('))
                    esteps = int(l[3].rstrip(')'))
            if 'FORCE on cell' in line:
                look_for_total = True
            if look_for_total and 'Total' in line:
                try:
                    stresses = [ ffloat(v) for v in line.split()[1:] ]
                except:
                    stresses = [ 
                            ffloat(line[9:20]), 
                            ffloat(line[21:32]), 
                            ffloat(line[33:44]),
                            ffloat(line[45:56]),
                            ffloat(line[57:68]),
                            ffloat(line[69:80])]
                look_for_total = False

        self.clean_fields()

        ### assign hubbard terms
        self.new_hubbards = []
        if ldauls and ldauus:
            self.ldau = True
            for elt, l, u in zip(species, ldauls, ldauus):
                hub, created = Hubbard.objects.get_or_create(
                        element_id=elt, ldau_u=u, ldau_l=l)
                if created:
                    hub.save()
                self.add_hubbard(hub)

        if esteps == self.nelm:
            self.econverged = False
        else:
            self.econverged = True

        if not self.voluntary:
            self.add_error('incomplete')
        elif not self.econverged and self.settings == 'standard':
            self.add_error('electronic_convergence')
        elif self.nsw > 1 and not self.fconverged:
            self.add_error('ionic_convergence')
        elif not self.get_errors():
            self.done = True

        ### create input structure
        if self.input is None:
            struct = Structure()
            struct.cell = input_cell
            for ind in range(len(atom_types)):
                atom = Atom()
                atom.element = atom_types[ind]
                atom.coord = input_coords[ind]
                struct.add_atom(atom)
            self.input = struct

        ### create output structure if applicable
        if self.niter > 1:
            output = Structure()
            output.cell = cell 
            output.stresses = stresses
            for ind in range(len(atom_types)):
                atom = Atom()
                atom.element = atom_types[ind]
                atom.direct = False
                atom.coord = coords[ind]
                atom.force = forces[ind]
                if magmoms: atom.magmom = magmoms[ind]
                if charges: atom.charge = charges[ind]
                output.add_atom(atom)
            self.output = output
        #else:
        #    self.output = self.input

        try:
            self.read_doscar()
        except:
            self.add_error('doscar_exc')
        self.pack()

    def read_doscar(self):
        if not os.path.exists(self.path+'/DOSCAR'):
            return None
        doscar = open(self.path+'/DOSCAR').readlines()
        bot = 0
        if len(doscar) < 20:
            return False
        for i in range(6):
            l=doscar.pop(0).rstrip()
        efermi = float(l.split()[3])
        self.fermi = efermi
        step1 = doscar.pop(0).split()[0]
        while step1 == '********':
            step1 = doscar.pop(0).split()[0]
        step2 = doscar.pop(0).split()[0]
        step_size = float(step2)-float(step1)
        not_found = True
        for l in doscar:
            l = l.split()
            if not not_found:
                break
            e = float(l.pop(0))
            dens = 0
            for i in range(len(l)/2):
                dens += float(l[i])
            if e < efermi and dens > 1e-3:
                bot = e
            elif e > efermi and dens > 1e-3:
                top = e
                not_found = False
            else:
                top = e
        if not bot:
            gap = 0
        else:
            if top - bot < step_size*2:
                gap = 0
            else:
                gap = top - bot
        if not not_found:
            self.band_gap = gap

    ### Error tracking stuff ###

    def add_error(self, error):
        errors = self.get_errors()
        if error not in errors:
            errors.append(error)
        self.errors = json.dumps(errors)

    def remove_error(self, error):
        errors = self.get_errors()
        if error in errors:
            errors.remove(error)
        self.errors = json.dumps(errors)

    def get_errors(self):
        return json.loads(self.errors)

    def check_stdout(self, filename='stdout.txt'):
        if not os.path.exists(self.path+'/'+filename):
            return
        stdout = open('%s/%s' % (self.path, filename)).read()
        if 'Error reading item' in stdout:
            self.add_error('input_error')
        if 'ZPOTRF' in stdout:
            self.add_error('zpotrf')
        if 'SGRCON' in stdout:
            self.add_error('sgrcon')
        if 'INVGRP' in stdout:
            self.add_error('invgrp')
        if 'BRIONS problems: POTIM should be increased' in stdout:
            self.add_error('brions')
        if 'FEXCF' in stdout:
            self.add_error('fexcf')
        if 'FEXCP' in stdout:
            self.add_error('fexcp')
        if 'PRICEL' in stdout:
            self.add_error('pricel')
        if 'EDDDAV' in stdout:
            self.add_error('edddav')
        if 'Sub-Space-Matrix is not hermitian in DAV' in stdout:
            self.add_error('hermitian')
        #if 'IBZKPT' in stdout:
        #    self.log += 'IBZKPT error'
        if 'BRMIX: very serious problems' in stdout:
            self.add_error('brmix')

    ### Problem solutions ###

    def address_errors(self):
        if not self.get_errors():
            print '   -Found no errors'
            return self
        
        new_calc = self.copy()
        errors = self.get_errors()
        
        #if errors == ['attempts']:
        #    new_calc.remove_error('attempts')
        #    return new_calc

        #if set(errors) == set(['ionic_convergence', 'attempts']):
        #    new_calc.remove_error('attempts')
        #    errors.remove('attempts')

        for err in errors:
            if err in ['duplicate','partial']:
                continue
            elif err == 'ionic_convergence':
                if self.niter > 1:
                    new_calc.remove_error('ionic_convergence')
                    new_calc.input = self.output.copy()
            elif err == 'electronic_convergence':
                new_calc.fix_electronic_convergence()
            elif err == 'doscar_exc':
                new_calc.fix_bands()
            elif err == 'bands':
                new_calc.fix_bands()
            elif err == 'edddav':
                new_calc.fix_dav()
            elif err == 'errrmm':
                new_calc.fix_rmm()
            elif err == 'sgrcon':
                new_calc.fix_sgrcon()
            elif err == 'zpotrf':
                new_calc.fix_zpotrf()
            elif err == 'invgrp':
                new_calc.fix_invgrp()
            elif err == 'exception':
                new_calc.clean_start()
            elif err == 'fexcf':
                new_calc.fix_fexcf()
            elif err == 'fexcp':
                new_calc.fix_fexcp()
            elif err == 'brions':
                new_calc.fix_brions()
            elif err == 'pricel':
                new_calc.fix_pricel()
            elif err == 'brmix':
                new_calc.fix_brmix()
            elif err == 'incomplete':
                continue
            elif err == 'hermitian':
                new_calc.fix_hermitian()
            else:
                print '   -Unsolvable error: %s' % err

        if 'incomplete' in new_calc.get_errors():
            if self.niter > 1:
                new_calc.input = self.output.copy()
                new_calc.remove_error('incomplete')
        if 'magnetic' in self.entry.keywords:
            new_calc.input.set_magnetism('ferro')
        return new_calc

    def move(self, path):
        path = os.path.abspath(path)
        try:
            os.system('mv %s %s' % (self.path, path))
        except Exception, err:
            print err
            return
        self.path = path
        self.save()

    def move_contents(self, path):
        path = os.path.abspath(path)
        try:
            os.system('mkdir %s &> /dev/null' % path)
            os.system('cp %s/* %s &> /dev/null' % (self.path, path))
            os.system('rm %s/* &> /dev/null' % self.path)
        except Exception, err:
            print err
            return
        self.path = path
        self.save()

    def backup(self):
        new_dir = '%s_' % self.attempt
        new_dir += '_'.join(self.get_errors())
        print '   -Backing up %s to %s' % (self.path, new_dir)
        self.move_contents(self.path+'/'+new_dir)

    def clean_start(self):
        depth = self.path.count('/') - self.path.count('..')
        if depth < 6:
            print '   -Too short path supplied to clean_start', self.path
            return
        else:
            print '   -Wiping path: %s' % self.path
            os.system('rm -rf %s &> /dev/null' % self.path)

    def fix_zhegev(self):
        raise NotImplementedError

    def fix_brmix(self):
        self.symprec = 1e-7
        self.algo = 'normal'
        self.remove_error('brmix')
        self.remove_error('incomplete')
        print '   -Fixed BRMIX error'

    def fix_fexcf(self):
        self.remove_error('fexcf')
        self.remove_error('incomplete')
        self.potim = 0.1
        self.algo = 'normal'
        print '   -Fixed FEXCF error'

    def fix_fexcp(self):
        self.remove_error('fexcp')
        self.remove_error('incomplete')
        self.potim = 0.1
        self.algo = 'normal'
        print '   -Fixed FEXCP error'

    def fix_pricel(self):
        self.remove_error('pricel')
        self.remove_error('incomplete')
        self.symprec=1e-7
        print '   -Fixed PRICEL error'

    def fix_electronic_convergence(self):
        if self.algo != 'normal':
            self.algo = 'normal'
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            print '   -switching to DAV'
        else:
            print '   -Can\'t fix electronic convergence.'

    def fix_invgrp(self):
        self.remove_error('invgrp')
        self.remove_error('incomplete')
        self.symprec=1e-7
        print '   -Fixed INVGRP error'

    def fix_sgrcon(self):
        self.remove_error('sgrcon')
        self.remove_error('incomplete')
        self.symprec=1e-7
        print '   -Fixed SGRCON error'

    def fix_brions(self):
        self.remove_error('brions')
        self.remove_error('incomplete')
        self.potim *= 2
        print '   -Fixed BRIONS problem'

    def fix_zpotrf(self):
        self.remove_error('zpotrf')
        self.remove_error('incomplete')
        self.algo = 'normal'
        self.potim = 0.1
        print '   -Fixed ZPOTRF error'

    def fix_bands(self):
        self.remove_error('bands')
        self.remove_error('incomplete')
        self.remove_error('doscar_exc')
        self.remove_error('electronic_convergence')
        self.nbands *= int(np.ceil(1.15))
        print '   -Fixed bands error: %s' % self.entry

    def fix_dav(self):
        if self.algo == 'fast':
            self.algo = 'very_fast'
            self.remove_error('edddav')
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            print '   -Error with DAV, switching to RMM'
        elif self.algo == 'normal':
            self.algo = 'very_fast'
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            self.remove_error('edddav')
            print '   -Problem with DAV and RMM. =/'
        elif self.algo == 'very_fast':
            print "   -Can't fix EDDDAV. Stopping."

    def fix_rmm(self):
        if self.algo == 'fast':
            self.algo = 'normal'
            self.remove_error('errrmm')
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            print '   -Error with RMM, switching to DAV'
        elif self.algo == 'very_fast':
            print '   -Problem with DAV and RMM. =/'
            self.algo = 'normal'
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            self.remove_error('errrmm')
        elif self.algo == 'normal':
            print "   -Can't fix RMM. Stopping."

    def fix_hermitian(self):
        if not self.algo == 'very_fast':
            self.algo = 'very_fast'
            self.remove_error('hermitian')
            self.remove_error('incomplete')
            self.remove_error('electronic_convergence')
            print "   -Fixing 'sub-space not hermitian' error."
        else:
            print "   -Can't fix 'sub-space not hermitian'. Stopping."

    #### calculation management

    def write(self):
        os.system('mkdir %s &> /dev/null' % self.path)
        poscar = open(self.path+'/POSCAR','w')
        potcar = open(self.path+'/POTCAR','w')
        incar = open(self.path+'/INCAR','w')
        kpoints = open(self.path+'/KPOINTS','w')
        poscar.write(self.POSCAR)
        potcar.write(self.POTCAR)
        incar.write(self.INCAR)
        kpoints.write(self.KPOINTS)
        poscar.close()
        potcar.close()
        incar.close()
        kpoints.close()

    @property
    def estimate(self):
        return 24*8*3600

    @property
    def instructions(self):
        if self.done:
            return {}
        elif self.get_errors():
            print '   -Still had errors:', self.get_errors()
            return {}

        instruction = {
                'path':self.path,
                'walltime':self.estimate,
                'header':'gunzip -f CHGCAR.gz &> /dev/null\ndate +%s\nulimit -s unlimited',
                'mpi':'mpirun -machinefile $PBS_NODEFILE -np $NPROCS',
                'binary':'vasp_53', 
                'pipes':' > stdout.txt 2> stderr.txt',
                'footer':'gzip -f CHGCAR OUTCAR PROCAR ELFCAR\nrm -f WAVECAR vasprun.xml\ndate +%s'}

        if self.input.natoms < 10:
            instruction.update({'mpi':'','binary':'vasp_53_serial',
                'serial':True})
        return instruction

    @staticmethod
    def do(input=None, type='standard', path=None, entry=None,
            args=[], hubbards='wang_act', potentials='oqmd_pbe', 
            **kwargs):
        
        t0 = time.time()
        # where to run
        if path is None:
            path = os.path.abspath(entry.path)
        else:
            path = os.path.abspath(path)

        print '   -Doing %s on %s!' % (type, entry)
        # is this an existing calculation in the given path?
        try:
            calc = Calculation.objects.get(path=path)
        except:
            # if the path doesn't exist
            if not os.path.exists(path):
                os.mkdir(path)

            calc = Calculation()
            calc.settings = type
            calc.path = path
            calc.input = input
            if any(a.magmom for a in calc.input.atoms):
                calc.ispin = 2
            calc.set_args(args)
            if entry:
                calc.entry=entry

        if calc.done:
            print 'Calc is done'
            return calc

        # 2
        ps = PotentialSet.objects.get(name=potentials)

        for elt1 in calc.input.elements:
            if elt1.induced_hubbards.filter(hubbardset__name=hubbards).exists():
                for elt2 in calc.input.elements:
                    hub, created = elt1.induced_hubbards.get_or_create(
                            element=elt2)
                    if created:
                        hub.save()
                    calc.add_hubbard(hub)

        for elt in calc.input.elements:
            calc.add_potential(ps.get(elt.symbol))

        calc.npar = 2
        if calc.input.natoms > 20:
            calc.lreal = 'auto'

        for k, v in vasp_settings[type].items():
            setattr(calc, k, v)

        if 'scale_encut' in vasp_settings:
            enmax = max(pot.enmax for pot in calc.potentials)
            calc.encut = int(vasp_settings['scale_encut']*enmax)

        if set(calc.get_errors()) <= set(['attempts', 'ionic_convergence']):
            calc.remove_error('attempts')
            calc.remove_error('ionic_convergence')

        # has the calculation been run yet?
        if not ( (os.path.exists(calc.path+'/OUTCAR') or 
            os.path.exists(calc.path+'/OUTCAR.gz')) and 
            os.path.exists(calc.path+'/CONTCAR')):
            ### but has it been written?
            if not ( os.path.exists(calc.path+'/INCAR') and
                    os.path.exists(calc.path+'/POTCAR') and 
                    os.path.exists(calc.path+'/KPOINTS') and
                    os.path.exists(calc.path+'/POSCAR')):
                calc.write()
            return calc

        calc.read_outcar()
        calc.check_stdout()

        if calc.done:
            return calc
        elif not calc.get_errors():
            calc.write()
            return calc
        
        fixed_calc = calc.address_errors()
        calc.backup()
        fixed_calc.clear_outputs()
        fixed_calc.write()
        return fixed_calc



from django.db import models

class Element(models.Model):
    ### Identification
    z = models.IntegerField(db_index=True)
    name = models.CharField(max_length=20)
    symbol = models.CharField(max_length=9, primary_key=True)

    ### Periodic table
    grp = models.IntegerField()
    period = models.IntegerField()

    ### Phyical characteristics
    mass = models.FloatField()
    density = models.FloatField()
    volume = models.FloatField()
    atomic_radii = models.IntegerField()
    van_der_waals_radii = models.IntegerField()
    covalent_radii = models.IntegerField()

    ### Thermodynamics
    melt = models.FloatField()
    boil = models.FloatField()
    specific_heat = models.FloatField()

    ### Electonic structure
    electronegativity = models.FloatField()
    first_ionization_energy = models.FloatField()
    s_elec = models.IntegerField()
    p_elec = models.IntegerField()
    d_elec = models.IntegerField()
    f_elec = models.IntegerField()

    ### misc
    abundance = models.FloatField()
    production = models.FloatField(default=0)
    radioactive = models.BooleanField(default=False)

    ### calculated properties
    base_mu = models.FloatField(null=True)

    class Meta:
        db_table = 'elements'

    def __str__(self):
        return self.symbol

    @property
    def get_elementstring(self):
        save_str = ''
        for k, v in self.fields.items():
            save_str += '{key} = {val}\n'.format(key=k, val=str(v))
        return save_str

    @staticmethod
    def write_elementfiles(dirname):
        import os
        os.chdir(dirname)

        elements = Element.objects.all()
        for elt in elements:
            f = open(elt.symbol,'w')
            f.write(elt.get_elementstring)
            f.close()

    @staticmethod
    def read_elementfiles(dirname):
        import os
        for elt in os.listdir(dirname):
            var_dict = {}
            for line in open(dirname+'/'+elt, 'r'):
                k, v = line.split(' = ')
                var_dict[k] = v.rstrip('\n')
            elt, created = Element.objects.get_or_create(**var_dict)
            if created:
                elt.save()

    @staticmethod
    def get(symbol):
        return Element.objects.select_related().get(symbol=symbol)

    @property
    def mu(self):
        all = self.entry_set.filter(ntypes=1,
                calculation__settings='standard',
                calculation__done=True)
        gs = sorted(all, key=lambda x: x.energy)
        if not gs:
            return None 
        return gs[0].energy
        
class ElementSet(models.Model):
    name = models.CharField(max_length=20, primary_key=True)
    elements = models.ManyToManyField(Element)
    class Meta:
        db_table = 'element_classes'

    @staticmethod
    def read(pgfile):
        e_string = open(pgfile).read()
        e_string = e_string.strip()
        egroup = {}
        elts = []
        for line in e_string.split('\n'):
            line = line.lstrip('-')
            if ' = ' in line:
                k, v = line.split(' = ')
                egroup[k] = v
            else:
                elts.append(line)
        eg, created = ElementSet.objects.get_or_create(**egroup)
        if not created:
            return eg
        eg.save()
        for elt in elts:
            eltobj = Element.objects.get(symbol=elt)
            eg.elements.add(eltobj)
        return eg

    #@property
    #def elements(self):
    #    return list(self.element_set.all())

class Potential(models.Model):
    potid = models.AutoField(primary_key=True)
    potcar = models.TextField()
    element = models.ForeignKey(Element)

    name = models.CharField(max_length=10)
    xc = models.CharField(max_length=3)
    gw = models.BooleanField(default=False)
    paw = models.BooleanField(default=False)
    us = models.BooleanField(default=False)
    enmax = models.FloatField()
    enmin = models.FloatField()
    date = models.CharField(max_length=20)
    electrons = models.TextField(blank=True, null=True)

    class Meta:
        db_table = 'potentials'

    def __str__(self):
        ident = '%s %s' % (self.name, self.xc)
        if self.paw:
            ident += ' PAW'
        if self.us:
            ident += ' US'
        if self.gw:
            ident += ' GW'
        return ident

    @staticmethod
    def read_potcar(potfile):
        pots = open(potfile).read()
        pots = pots.strip().split('End of Dataset')
        potobjs = []
        for pot in pots:
            if not pot:
                continue
            potcar = {}
            for line in pot.split('\n'):
                if 'TITEL' in line:
                    potcar['name'] = line.split()[3]
                    elt = potcar['name'].split('_')[0]
                    date = potcar['name'].split('_')[-1]
                    potcar['element'] = Element.objects.get(symbol=elt)
                    if 'GW' in line:
                        potcar['gw'] = True
                    if 'PAW' in line:
                        potcar['paw'] = True
                    if 'US' in line:
                        potcar['us'] = True
                if 'ENMAX' in line:
                    data = line.split()
                    potcar['enmax'] = float(data[2].rstrip(';'))
                    potcar['enmin'] = float(data[5])
                if 'VRHFIN' in line:
                    potcar['electrons'] = line.split(':')[1]
                if 'LEXCH' in line:
                    key = line.split()[-1]
                    if key == '91':
                        potcar['xc'] = 'GGA'
                    elif key == 'CA':
                        potcar['xc'] = 'LDA'
                    elif key == 'PE':
                        potcar['xc'] = 'PBE'
            potobj, created = Potential.objects.get_or_create(**potcar)
            if created:
                potobj.potcar = pot
            potobjs.append(potobj)
        return potobjs

class Hubbard(models.Model):
    element = models.ForeignKey(Element, related_name='hubbards')
    with_elt = models.ForeignKey(Element, related_name='induced_hubbards',
            null=True, blank=True, default=None)
    ldau_u = models.FloatField(default=0)
    ldau_l = models.IntegerField(default=-1)
    class Meta:
        db_table = 'hubbards'

    def __str__(self):
        if self.with_elt is not None:
            return '%s (with %s) U=%s L=%s' % (self.element.symbol,
                    self.with_elt.symbol, self.ldau_u, self.ldau_l)
        else:
            return '%s U=%s L=%s' % (self.element.symbol, 
                    self.ldau_u, self.ldau_l)


    def __nonzero__(self):
        if self.ldau_u > 0 and self.ldau_l != -1:
            return True
        else:
            return False

    def __eq__(self, other):
        if self.element != other.element:
            return False
        if self.ldau_u != other.ldau_u:
            return False
        if self.ldau_l != other.ldau_l:
            return False
        return True

    @property
    def get_hubbardstring(self):
        return str(self)

def read_potentials():
    import os
    potdir = '/home/sjk648/bin/qmpy/config/potentials'
    dirs = os.listdir(potdir) 
    for d in dirs:
        for pot in os.listdir(potdir+'/'+d):
            if not os.path.exists(potdir+'/'+d+'/'+pot+'/POTCAR'):
                print 'No', d+'/'+pot+'/POTCAR'
                continue
            potobj = Potential.read_potcar(potdir+'/'+d+'/'+pot+'/POTCAR')[0]
            potobj.save()

def read_elements():
    Element.read_elementfiles('/home/sjk648/bin/qmpy/config/elements')


class PotentialSet(models.Model):
    name = models.CharField(max_length=30)
    xc = models.CharField(max_length=3, default='PBE')
    gw = models.BooleanField(default=False)
    paw = models.BooleanField(default=True)
    us = models.BooleanField(default=False)
    potentials = models.ManyToManyField(Potential)

    class Meta:
        db_table = 'potential_sets'

    def get(self, element):
        return self.potentials.get(element__symbol=element)

    def get_pg_str(self):
        outstr = 'name = %s\n' % self.name
        outstr += '-xc = %s\n' % self.xc 
        outstr += '-paw = %s\n' % self.paw
        outstr += '-us = %s\n' % self.us
        for pot in self.potentials.all():
            outstr += '--%s\n' % pot.name

    @property
    def filter(self):
        return {'us':self.us, 
                'gw':self.gw,
                'xc':self.xc,
                'paw':self.paw}

    @staticmethod
    def read(pgfile):
        pg_string = open(pgfile).read()
        pg_string = pg_string.rstrip('\n')
        potgroup = {}
        pots = []
        for line in pg_string.split('\n'):
            line = line.lstrip('-')
            if ' = ' in line:
                k, v = line.split(' = ')
                potgroup[k] = v
            else:
                pots.append(line)
        pg, created = PotentialSet.objects.get_or_create(**potgroup)
        if not created:
            return pg
        pg.save()
        for pot in pots:
            potobj = Potential.objects.filter(**pg.filter).get(name=pot)
            pg.potentials.add(potobj)
        return pg

    def __str__(self):
        return "%s : %s" % (self.name, self.xc)


class HubbardSet(models.Model):
    name = models.CharField(max_length=30)
    hubbards = models.ManyToManyField(Hubbard)
    class Meta:
        db_table = 'hubbard_sets'

    def get(self, element):
        hub, created = self.hubbards.get_or_create(element__symbol=element)
        if created:
            hub.save()
        return hub

    def get_pg_str(self):
        for hub in self.hubbards.all():
            outstr += hubbard.get_hubbardstring()

    @staticmethod
    def read(path):
        h_string = open(path).read()
        h_string = h_string.strip()
        for line in h_string.split('\n'):
            if ' = ' in line:
                line = line.split(' = ')
                hub_group, created = HubbardSet.objects.get_or_create(name=line[1])
                if not created:
                    return hub_group
                continue
            line = line.split()
            hubbard, created = Hubbard.objects.get_or_create(element_id=line[0], 
                    with_elt_id=line[1],
                    ldau_u=float(line[2]), ldau_l=float(line[3]))
            hub_group.hubbards.add(hubbard)
        return hub_group

    def get_hubbardgroup_string(self):
        hstring = ''
        for h in self.hubbards.all():
            hstring += h.get_hubbardstring()+'\n'
        return hstring

    def __str__(self):
        return self.name

def read_hubbard_groups():
    import os
    path='/home/sjk648/bin/qmpy/config/collections/hubbards'
    groups = os.listdir(path)
    HubbardSet.objects.all().delete()
    for group in groups:
        hg = HubbardSet.read(path+'/'+group)
        hg.save()

def read_element_groups():
    import os
    path='/home/sjk648/bin/qmpy/config/collections/elements'
    groups = os.listdir(path)
    ElementSet.objects.all().delete()
    for group in groups:
        eg = ElementSet.read(path+'/'+group)
        eg.save()

def read_potential_groups():
    import os
    path='/home/sjk648/bin/qmpy/config/collections/potentials'
    groups = os.listdir(path)
    PotentialSet.objects.all().delete()
    for group in groups:
        pg = PotentialSet.read(path+'/'+group)
        pg.save()



from qmpy import *
from datetime import datetime, timedelta
import json
import os.path
import time
import sys

def check_die():
    if os.path.exists('/home/sjk648/die'):
        exit(0)

def collect_jobs():
    print 'collecting all jobs'
    for j in Job.objects.filter(state=1):
        print j
        check_die()
        j.collect()
        j.state = 2
        j.save()

def check_jobs():
    while True:
        jobs = Job.objects.filter(state=1, account__host__state=1)
        for job in jobs:    
            check_die()
            if job.is_done():
                print job
                job.collect()
                job.state = 2
                job.save()
        for i in range(20):
            time.sleep(1)
            check_die()


def check_tasks(project=None, path=None, id=None):
    while True:
        check_die()

        tasks = Task.objects.filter(state=0, 
                project__state=1,
                entry__holds='')
        if project:
            tasks = tasks.filter(project=project)
        if path:
            tasks = tasks.filter(entry__path__contains=path)
        if id:
            tasks = tasks.filter(id__endswith=id)
        tasks = tasks.order_by('priority')
        if project == 'questek':
            tasks = tasks.order_by('entry__path')
        #if project == 'icsd':
        #    tasks = tasks.filter(entry__natoms__lt=10)

        while not tasks.exists():
            print 'No tasks available'
            Project.objects.filter(state=0).update(state=1)
            for i in range(120):
                time.sleep(1)
                check_die()

            tasks = Task.objects.filter(state=0, 
                    project__state=1,
                    entry__holds='')
            if project:
                tasks = tasks.filter(project=project)
            if path:
                tasks = tasks.filter(entry__path__contains=path)
            if id:
                tasks = tasks.filter(id__endswith=id)
            tasks = tasks.order_by('priority')
            if project == 'questek':
                tasks = tasks.order_by('entry__path')
            #if project == 'icsd':
            #    tasks = tasks.filter(entry__natoms__lt=10)

        task = tasks[0]
        print task
        if not task.eligible_to_run:
            continue

        try:
            jobs = task.get_jobs()
        except VaspError:
            print '    -VASP Error.'
            task.state = -1
            task.save()
            continue
        except Exception, err:
            task.state = -1
            task.save()
            print '   -Unknown error!', err
            continue

        for job in jobs:
            print '   -Submitting', job
            while not job.state == 1:
                #try:
                job.submit()
                #except Exception:
                #    print '    -submission error, waiting 30 seconds'
                #    time.sleep(30)
            job.save()
        task.save()

if __name__ == '__main__':
    if 'force' in sys.argv:
        collect_jobs()
    if 'collect' in sys.argv:
        check_jobs()
    if 'submit' in sys.argv:
        try:
            project = sys.argv.index('project')
            project = sys.argv[project+1]
        except ValueError:
            project = None

        try:
            path = sys.argv.index('path')
            path = sys.argv[path+1]
        except ValueError:
            path = None

        try:
            id = sys.argv.index('id')
            id = sys.argv[id+1]
        except ValueError:
            id = None

        check_tasks(project=project, path=path, id=id)



from numpy.linalg import solve
import time
import copy
from collections import defaultdict

from qmpy.utils import *
from dev_qmpy.models.structure import Structure, Atom
from dev_qmpy.models.element import Element

def to_vasp(struct, direct=True):
    comp = struct.comp
    ordered_keys = sorted(comp.keys())
    atoms = sorted(struct.atoms, key=lambda x:x.element.symbol)
    poscar = ' '.join(set(a.symbol for a in atoms)) + '\n 1.0\n'
    cell = '\n'.join([ ' '.join([ '%f' % v  for v in vec ]) for vec in
        struct.cell ])
    poscar += cell +'\n'
    names = ' '.join( a for a in ordered_keys ) + '\n'
    ntypes = ' '.join( str(comp[k]) for k in ordered_keys ) + '\n'
    poscar += names
    poscar += ntypes
    if direct:
        poscar += 'direct\n'
        for x,y,z in struct.scaled_coords:
            poscar += ' %.10f %.10f %.10f\n' % (x,y,z)
    else:
        poscar += 'cartesian\n'
        for x, y, z in struct.coords:
            poscar += ' %.10f %.10f %.10f\n' % (x,y,z)
    return poscar

def from_vasp(poscar):
    struct = Structure()
    poscar = open(poscar,'r')
    title = poscar.readline().strip()
    scale = float(poscar.readline().strip())
    s = float(scale)
    cell = [[ float(v) for v in poscar.readline().split() ],
            [ float(v) for v in poscar.readline().split() ],
            [ float(v) for v in poscar.readline().split() ]]
    cell = np.array(cell)

    if s > 0:
        struct.cell = cell*s
    else:
        raise NotImplementedError, 'Not able to scale volume like this yet'

    vasp5 = False
    species = poscar.readline().strip().split()
    try:
        float(species[0])
    except:
        vasp5 = True
        counts = [ int(v) for v in poscar.readline().split() ]

    if not vasp5:
        species = title.strip().split()
        counts = species
        print 'Since not vasp5 format, title MUST be species present'

    atom_types = []
    for n,e in zip(counts, species):
        atom_types += [e]*n

    style = poscar.readline()
    direct = False
    if style[0] in ['D', 'd']:
        direct = True

    natoms = sum(counts)
    for i in range(natoms):
        atom = Atom()
        atom.element = Element.get(atom_types[i])
        atom.coord = [ float(v) for v in poscar.readline().split() ]
        atom.direct = direct
        struct.add_atom(atom)
    return struct

def to_incar(struct):
    incar = ''
    if any(a.magmom for a in struct.atoms):
        incar += ' ISPIN = 2\n'
        moms = [ a.magmom for a in struct.atoms ]
        magmoms = [[1, moms[0]]]
        for n in range(1, len(moms)):
            if moms[n] == moms[n-1]:
                magmoms[-1][0] += 1
            else:
                magmoms.append([1, moms[n]])
        momstr = ' '.join('%i*%.4f' % (v[0],v[1]) for v in magmoms)
        incar += ' MAGMOM = %s\n' %  momstr
    return incar



from django.db import models
from numpy.linalg import solve
import time
import copy
from pyspglib import spglib
from collections import defaultdict

from dev_qmpy.models.composition import Composition
from dev_qmpy.models.element import Element
from dev_qmpy.utils import *


class Atom(models.Model):
    structure = models.ForeignKey('Structure', related_name='atom_set')
    element = models.ForeignKey('Element', related_name='atoms')

    x = models.FloatField(blank=False)
    y = models.FloatField(blank=False)
    z = models.FloatField(blank=False)
    direct = models.BooleanField(default=True)

    fx = models.FloatField(default=0)
    fy = models.FloatField(default=0)
    fz = models.FloatField(default=0)

    magmom = models.FloatField(default=0)
    occupancy = models.FloatField(default=1)

    charge = models.FloatField(blank=True, null=True)
    volume = models.FloatField(blank=True, null=True)

    class Meta:
        db_table = 'atoms'

    def __str__(self):
        return '%s: %8f %8f %8f' % (self.element.symbol, 
                self.x, self.y, self.z)

    @property
    def symbol(self):
        return self.element.symbol

    @symbol.setter
    def symbol(self, symbol):
        self.element = Element.get(symbol)

    @property
    def force(self):
        return np.array([self.fx , self.fy, self.fz])

    @force.setter
    def force(self, values):
        self.fx, self.fy, self.fz = values

    @property
    def coord(self):
        return np.array([self.x , self.y, self.z])

    @coord.setter
    def coord(self, values):
        self.x, self.y, self.z = values

class Structure(models.Model, object):
    id = models.AutoField(primary_key=True)
    entry = models.ForeignKey('Entry', null=True)
    label = models.CharField(blank=True, max_length=20)
    log = models.TextField()

    ## representations
    composition = models.ForeignKey(Composition, null=True)
    name = models.CharField(max_length=255, db_index=True, null=True)
    formula = models.CharField(max_length=255, db_index=True, null=True)
    natoms = models.IntegerField(null=True)

    x1 = models.FloatField()
    x2 = models.FloatField()
    x3 = models.FloatField()
    y1 = models.FloatField()
    y2 = models.FloatField()
    y3 = models.FloatField()
    z1 = models.FloatField()
    z2 = models.FloatField()
    z3 = models.FloatField()

    ## stresses
    sxx = models.FloatField(default=0)
    syy = models.FloatField(default=0)
    szz = models.FloatField(default=0)
    sxy = models.FloatField(default=0)
    syz = models.FloatField(default=0)
    szx = models.FloatField(default=0)

    ## symmetry
    spacegroup = models.IntegerField(null=True)

    class Meta:
        db_table = 'structures'

    def __eq__(self, other):
        if self.comp != other.comp:
            return False
        if not np.allclose(self.cell, other.cell):
            return False
        if not np.allclose(self.scaled_coords, other.scaled_coords):
            return False
        return True

    def __str__(self):
        return comp_to_name(self.comp, special='reduce')

    def like(self, other):
        if self.spacegroup is None:
            self.find_symmetry()
            self.save()
        if other.spacegroup is None:
            other.find_symmetry()
            other.save()

    def save(self,*args, **kwargs):
        self.formula = comp_to_formula(self.comp)
        self.name = comp_to_name(self.comp, special='reduce')
        self.composition = Composition.get(self.comp)
        if hasattr(self, 'new_atoms'):
            self.natoms = len(self.atoms)
        else:
            self.natoms = self.atom_set.count()
        super(Structure, self).save(*args, **kwargs)
        if hasattr(self, 'new_atoms'):
            for atom in self.new_atoms:
                self.atom_set.add(atom)
            del self.new_atoms

    @property
    def volume(self):
        b1, b2, b3 = self.cell
        return abs(np.dot(np.cross(b1, b2), b3))

    @volume.setter
    def volume(self, value):
        scale = value/self.volume
        self.cell = self.cell * scale**(1/3.)

    @property
    def lat_params(self):
        v1, v2, v3 = self.cell
        a = norm(v1)
        b = norm(v2)
        c = norm(v3)
        alpha = np.arccos(np.dot(v2, v3)/(b*c))*180/np.pi
        beta = np.arccos(np.dot(v1, v3)/(a*c))*180/np.pi
        gamma = np.arccos(np.dot(v1, v2)/(a*b))*180/np.pi
        return a, b, c, alpha, beta, gamma

    @property
    def cell(self):
        return np.array([
                [self.x1, self.x2, self.x3],
                [self.y1, self.y2, self.y3],
                [self.z1, self.z2, self.z3]])

    @cell.setter
    def cell(self, cell):
        self.x1, self.x2, self.x3 = cell[0]
        self.y1, self.y2, self.y3 = cell[1]
        self.z1, self.z2, self.z3 = cell[2]


    @property
    def atomic_numbers(self):
        return np.array([ atom.element.z for atom in self.atoms ])

    @property
    def stresses(self):
        return np.array([self.sxx, self.syy, self.szz, 
            self.sxy, self.syz, self.szx ])

    @stresses.setter
    def stresses(self, vector):
        self.sxx, self.syy, self.szz = vector[0:3]
        self.sxy, self.syx, self.szx = vector[3:6]

    @property
    def atoms(self):
        atoms = []
        if hasattr(self, 'new_atoms'):
            for a in self.new_atoms:
                atoms.append(a)
            return sorted(atoms, key=lambda x: x.symbol)
        return sorted(self.atom_set.all(), key=lambda x: x.symbol)

    def add_atom(self, atom):
        if not hasattr(self, 'new_atoms'):
            self.new_atoms = []
        self.new_atoms.append(atom)

    def set_magnetism(self, type, scheme='primitive'):
        if type == 'none':
            for atom in self.atoms:
                atom.magmom = 0
                if atom.id is not None:
                    atom.save()
        if type == 'ferro':
            for atom in self.atoms:
                if atom.element.d_elec > 0 and atom.element.d_elec < 10:
                    atom.magmom = 5
                elif atom.element.f_elec > 0 and atom.element.f_elec < 14:
                    atom.magmom = 7
                else:
                    atom.magmom = 0
                if atom.id is not None:
                    atom.save()
        elif type == 'anti-ferro':
            raise NotImplementedError

    @property
    def ntypes(self):
        return len(self.elements)

    @property
    def mrp(self):
        v1,v2,v3 = self.cell
        return  min(
                norm(v1),norm(v2),norm(v3),
                norm(v1+v2),norm(v2+v3),norm(v3+v1),
                norm(v1+v2+v3))

    @property
    def magmom(self):
        return sum(a.magmom for a in self.atoms)

    @property
    def elements(self):
        elts = set( atom.element for atom in self.atoms )
        return elts

    @property
    def comp(self):
        comp = defaultdict(int)
        for atom in self.atoms:
            comp[atom.symbol] += 1
        return comp

    @property
    def unit_comp(self):
        return defaultdict(float,
                name_to_comp(self.formula, special='unit'))

    @property
    def coords(self):
        return np.array([ atom.coord for atom in self.atoms ])

    @property
    def magmoms(self):
        return np.array([ atom.magmom for atom in self.atoms ])

    @property
    def scaled_coords(self):
        inv = self.inv.T
        coords = []
        for atom in self.atoms:
            if not atom.direct:
                coords.append(np.dot(inv, atom.coord))
            else:
                coords.append(atom.coord)
        coords = np.array(coords)
        coords %= 1.0
        coords %= 1.0
        return coords

    @property
    def cartesian_coords(self):
        coords = []
        for atom in self.atoms:
            if atom.direct:
                coords.append(np.dot(atom.coord, self.cell))
            else:
                coords.append(atom.coord)
        return np.array(coords)

    @property
    def forces(self):
        forces = []
        for atom in self.atoms:
            forces.append(atom.force)
        return np.array(forces)

    @forces.setter
    def forces(self, forces):
        for force, atom in zip(forces, self.atoms):
            atom.force = force

    @property
    def reciprocal_lattice(self):
        return np.linalg.inv(self.cell).T

    @property
    def inv(self):
        return np.linalg.inv(self.cell)

    @property
    def relative_rec_lat(self):
        rec_lat = self.reciprocal_lattice
        rec_mags = [ norm(rec_lat[0]), norm(rec_lat[1]), norm(rec_lat[2])]
        r0 = min(rec_mags)
        return np.array([ np.round(r/r0, 4) for r in rec_mags ])

    @property
    def sg(self):
        if not self.spacegroup:
            self.find_symmetry(keep=True)
        return self.spacegroup

    def find_symmetry(self, keep=False):
        sg = spglib.get_spacegroup(self.ase, symprec=1e-2)
        sg = sg.strip(' ()')
        self.spacegroup = int(sg)
        if keep:
            self.save()
        return int(sg)

    def get_distance(self, atom1, atom2):
        if isinstance(atom1, int):
            atom1 = self.atoms[atom1]
        if isinstance(atom2, int):
            atom2 = self.atoms[atom2]
        if atom1.direct:
            v1 = np.dot(atom1.coord, self.cell)
        else:
            v1 = atom1.coord
        if atom2.direct:
            v2 = np.dot(atom2.coord, self.cell)
        else:
            v2 = atom2.coord
        vec = v1 - v2
        Dr = solve(self.cell.T, vec)
        D = np.dot(Dr - np.round(Dr), self.cell)
        return norm(D)


    def find_neighbors(self):
        ''' Based on the following algorithm defines a list of pairs of atoms,
        sotred as a tuple, of the form (atom1,atom2,dist_1_2). For all triplets
        of atoms A, B and C, the atoms A and B are neighbors if and only if:
            AB is shorter than AC
            and
            AB is shorter than BC'''
        self.neighbors = []
        self.neigh_dict = defaultdict(list)
        atoms = self.atoms
        mrp = self.mrp
        for atom1, atom2 in itertools.combinations(atoms, 2):
            _12 = self.get_distance(atom1, atom2)
            if _12 == 0:
                _12 = mrp
            for atom3 in atoms:
                _23 = self.get_distance(atom2, atom3)
                _31 = self.get_distance(atom3, atom1)
                if _23 == 0: _23 = mrp
                if _31 == 0: _31 = mrp
                if _23 < _12 and _31 < _12:
                    break
            else:
                self.neighbors.append((atom1,atom2,_12))
                self.neigh_dict[atom1].append(atom2)
                self.neigh_dict[atom2].append(atom1)

    def substitute(self, replace, rescale=True):
        '''Replace atoms, as specified in a dict of pairs. 
        Optionally, can rescale the new volume to the total volume of the
        elemental phases of the new composition.

        Example:
        >>> s = Structure.create('POSCAR-Fe2O3')
        >>> s.substitute({'Fe':'Ni', 'O':'F'} rescale=True)
        '''

        struct = Structure()
        struct.cell = self.cell
        volume = 0.0
        for a, coord in zip(self.atoms, self.scaled_coords):
            a2 = Atom()
            a2.force = a.force
            a2.coord = coord
            a2.direct = True
            if a.symbol in replace:
                a2.symbol = replace[a.symbol]
            else:
                a2.symbol = a.symbol
            struct.add_atom(a2) 
            volume += a2.element.volume
        if rescale:
            struct.volume = volume
        return struct

    def copy(self):
        new = copy.deepcopy(self)
        new.id = None
        new.label = ''
        new.log = ''
        new.atom_set = []
        new.new_atoms = []
        for atom in self.atoms:
            new_atom = copy.deepcopy(atom)
            new_atom.id = None
            new.add_atom(new_atom)
        return new

    @property
    def similar(self):
        return Structure.objects.filter(natoms=self.natoms,
                composition=self.composition, 
                spacegroup=self.spacegroup,
                label=self.label)



# qmpy/base/query.py
from qmpy import *
import itertools
import numpy as np
from numpy.linalg import norm
import os
import sys
import random
import pprint
import re
import gzip
from collections import defaultdict
from my_strings import *

set_operators = {
        'and':'and',
        '&&':'and',
        'or':'or',
        '|':'or',
        'not':'not'}

search_operators = {
        '<':'lt','lt':'lt',
        '<=':'lte','lte':'lte',
        '>':'gt','gt':'gt',
        '>=':'gte','gte':'gte',
        '==':'exact','=':'exact',
        '!=':'ne','<>':'ne','ne':'ne',
        'contains':'contains'}

def get_model(model):
    model = model[0].upper() + model[1:].lower()
    try:
        model = getattr(sys.modules[__name__], model)
    except:
        return None
    if issubclass(model, models.Model):
        return model
    else:
        return None

def query(type='entry',
        formula='',
        search='',
        result='summary',
        skip_comp=False,
        unique=False):
    '''
    Most important function here. Governs the ability to give a search and get
    back a set of matching structures. This matching set is used as an argument
    of nearly every other function.
    '''

    model = get_model(type)
    result = model.objects.all()
    if not formula and not search:
        print 'No search criteria given, returning nothing'
        return result

    # Search on formula
    if formula:
        print formula
        formulae = parse_formula(formula)
        if any( 0 in d.values() for d in formulae):
            skip_comp = True
        f_result = Entry.objects.none()
        for comp in formulae:
            print comp
            if not skip_comp:
                name = comp_to_name(comp, special='reduce')
                if not Composition.objects.filter(name=name).exists():
                    tmpr = Entry.objects.none()
                    continue
                tmpr = Composition.objects.get(name=name).entry_set.all()
                f_result = f_result | tmpr
                continue
            ntypes = len(comp)
            tmpr = Entry.objects.all()
            for elt in comp:
                ents = Element.objects.get(symbol=elt).entry_set.all()
                nents = ents.filter(ntypes=ntypes)
                tmpr = tmpr & nents
            f_result = f_result | tmpr
        
        result = result & f_result

    # Search on traditional search criteria
    if search:
        if isinstance(search,str) or isinstance(search,unicode):
            search = search.split()
        next = 'column'
        chain = 'and'
        for s in search:
            if next == 'column':
                gen = parse_columns(s)
                next = 'operator'
            elif next == 'operator':
                operator = search_operators[s]
                next = 'value'
            elif next == 'value':
                value = s
                next = 'evaluate'
            elif next == 'chain':
                chain = set_operators[s]
                next = 'column'
            if next == 'evaluate':
                db,col = gen[0]
                if db == 'master':
                    if operator != 'ne':
                        tmpr = Entry.objects.filter(**{
                            '%s__%s' % (col,operator):value})
                    else:
                        tmpr = Entry.objects.exclude(**{
                            col:value})
                else:
                    if operator != 'ne':
                        tmpr = Entry.objects.filter(**{
                            str('%s__%s__%s' % (db,col,operator)):value})
                    else:
                        tmpr = Entry.objects.exclude(**{
                            str('%s__%s' % (db,col)):value})
                if chain == 'and':
                    result = result & tmpr
                elif chain == 'or':
                    result = result | tmpr
                elif chain == 'not':
                    result = result.exclude(pk__in=tmpr)
                next = 'chain'

    if unique:
        new_result = {}
        for r in result:
            s = cS.Structure(r.id)
            try:
                s.set_thermo('econverge')
            except:
                continue
            if s.name not in new_result:
                new_result[s.name] = [r,s]
            else:
                if s.dE < new_result[s.name][1].dE:
                    new_result[s.name] = [r,s]
        result = [ r[0] for r in new_result.values() ]

    if not result:
        print 'Nothing found!'
        return result
    else:
        return result




"""Module to read and write atoms in cif file format.

See http://www.iucr.org/resources/cif/spec/version1.1/cifsyntax for a
description of the file format.  STAR extensions as save frames,
global blocks, nested loops and multi-data values are not supported.
"""

import shlex
import re

import numpy as np

from ase.parallel import paropen
from ase.lattice.spacegroup import crystal
from ase.lattice.spacegroup.spacegroup import spacegroup_from_data


def get_lineno(fileobj):
    """Returns the line number of current line in fileobj."""
    pos = fileobj.tell()
    try:
        fileobj.seek(0)
        s = fileobj.read(pos)
        lineno = s.count('\n') 
    finally:
        fileobj.seek(pos)
    return lineno


def unread_line(fileobj):
    """Unread the last line read from *fileobj*."""

    # If previous line ends with CRLF, we have to back up one extra 
    # character before entering the loop below
    if fileobj.tell() > 2:
        fileobj.seek(-2, 1)
        if fileobj.read(2) == '\r\n':
            fileobj.seek(-1, 1)

    while True:
        if fileobj.tell() == 0:
            break
        fileobj.seek(-2, 1)
        if fileobj.read(1) in ('\n', '\r'):
            break
        

def convert_value(value):
    """Convert CIF value string to corresponding python type."""
    value = value.strip()
    if re.match('(".*")|(\'.*\')$', value):
        return value[1:-1]
    elif re.match(r'[+-]?\d+$', value):
        return int(value)
    elif re.match(r'[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?$', value):
        return float(value)
    elif re.match(r'[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?\(\d+\)$', 
                  value):
        return float(value[:value.index('(')])  # strip off uncertainties
    else:
        return value


def parse_multiline_string(fileobj, line):
    """Parse semicolon-enclosed multiline string and return it."""
    assert line[0] == ';'
    lines = [line[1:].lstrip()]
    while True:
        line = fileobj.readline().strip()
        if line.startswith(';'):
            break
        lines.append(line)
    return ' '.join(lines).strip()


def parse_singletag(fileobj, line):
    """Parse a CIF tag (entries starting with underscore). Returns
    a key-value pair."""
    kv = line.split(None, 1)
    if len(kv) == 1:
        key = line
        line = fileobj.readline().strip()
        while not line or line[0] == '#':
            line = fileobj.readline().strip()
        if line[0] == ';':
            value = parse_multiline_string(fileobj, line)
        else:
            value = line
    else:
        key, value = kv
    return key, convert_value(value)


def parse_loop(fileobj):
    """Parse a CIF loop. Returns a dict with column tag names as keys
    and a lists of the column content as values."""
    header = []
    line = fileobj.readline().strip()
    while line.startswith('_'):
        header.append(line.lower())
        line = fileobj.readline().strip()
    columns = dict([(h, []) for h in header])

    tokens = []
    while True:
        lowerline = line.lower()
        if (not line or 
            line.startswith('_') or 
            lowerline.startswith('data_') or 
            lowerline.startswith('loop_')):
            break
        if line.startswith('#'):
            line = fileobj.readline().strip()
            continue
        if line.startswith(';'):
            t = [parse_multiline_string(fileobj, line)]
        else:
            t = shlex.split(line)

        line = fileobj.readline().strip()

        tokens.extend(t)
        if len(tokens) < len(columns):
            continue
        assert len(tokens) == len(header), '\n%s\n%s' % (
                ' '.join(tokens), ' '.join(header))

        for h, t in zip(header, tokens):
            columns[h].append(convert_value(t))
        tokens = []
    if line:
        unread_line(fileobj)
    return columns


def parse_items(fileobj, line):
    """Parse a CIF data items and return a dict with all tags."""
    tags = {}
    while True:
        line = fileobj.readline()
        if not line:
            break
        line = line.strip()
        lowerline = line.lower()
        if not line or line.startswith('#'):
            continue
        elif line.startswith('_'):
            key, value = parse_singletag(fileobj, line)
            tags[key.lower()] = value
        elif lowerline.startswith('loop_'):
            tags.update(parse_loop(fileobj))
        elif lowerline.startswith('data_'):
            unread_line(fileobj)
            break
        elif line.startswith(';'):
            temp = parse_multiline_string(fileobj, line)
        else:
            raise ValueError('%s:%d: Unexpected CIF file entry: "%s"'%(
                    fileobj, get_lineno(fileobj), line))
    return tags


def parse_block(fileobj, line):
    """Parse a CIF data block and return a tuple with the block name
    and a dict with all tags."""
    assert line.lower().startswith('data_')
    blockname = line.split('_', 1)[1].rstrip()
    tags = parse_items(fileobj, line)
    return blockname, tags


def parse_cif(fileobj):
    """Parse a CIF file. Returns a list of blockname and tag
    pairs. All tag names are converted to lower case."""
    if isinstance(fileobj, basestring):
        fileobj = open(fileobj)

    import StringIO
    trans_table = ''.join( [chr(i) for i in range(128)] + [' '] * 128 )
    data = fileobj.read().translate(trans_table)
    fileobj = StringIO.StringIO()
    fileobj.write(data)
    fileobj.seek(0)

    blocks = []
    while True:
        line = fileobj.readline()
        if not line:
            break
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        blocks.append(parse_block(fileobj, line))
    return blocks


def tags2atoms(tags, store_tags=False, **kwargs):
    """Returns an Atoms object from a cif tags dictionary.  If
    *store_tags* is true, the *info* attribute of the returned Atoms
    object will be populated with all the cif tags.  Keyword arguments
    are passed to the Atoms constructor."""
    a = tags['_cell_length_a']
    b = tags['_cell_length_b']
    c = tags['_cell_length_c']
    alpha = tags['_cell_angle_alpha']
    beta = tags['_cell_angle_beta']
    gamma = tags['_cell_angle_gamma']

    scaled_positions = np.array([tags['_atom_site_fract_x'], 
                                 tags['_atom_site_fract_y'], 
                                 tags['_atom_site_fract_z']]).T

    symbols = []
    if '_atom_site_type_symbol' in tags:
        labels = tags['_atom_site_type_symbol']
    else:
        labels = tags['_atom_site_label']
    for s in labels:
        # Strip off additional labeling on chemical symbols
        m = re.search(r'([A-Z][a-z]?)', s)  
        symbol = m.group(0)
        symbols.append(symbol)

    # Symmetry specification, see
    # http://www.iucr.org/resources/cif/dictionaries/cif_sym for a
    # complete list of official keys.  In addition we also try to
    # support some commonly used depricated notations
    no = None
    if '_space_group.it_number' in tags:
        no = tags['_space_group.it_number']
    elif '_space_group_it_number' in tags:
        no = tags['_space_group_it_number']
    elif '_symmetry_int_tables_number' in tags:
        no = tags['_symmetry_int_tables_number']

    symbolHM = None
    if '_space_group.Patterson_name_h-m' in tags:
        symbolHM = tags['_space_group.patterson_name_h-m']
    elif '_symmetry_space_group_name_h-m' in tags:
        symbolHM = tags['_symmetry_space_group_name_h-m']

    sitesym = None
    if '_space_group_symop.operation_xyz' in tags:
        sitesym = tags['_space_group_symop.operation_xyz']
    elif '_symmetry_equiv_pos_as_xyz' in tags:
        sitesym = tags['_symmetry_equiv_pos_as_xyz']
        
    spacegroup = 1
    if sitesym is not None:
        spacegroup = spacegroup_from_data(no=no, symbol=symbolHM,
                                          sitesym=sitesym)
    elif no is not None:
        spacegroup = no
    elif symbolHM is not None:
        spacegroup = symbolHM
    else:
        spacegroup = 1

    if store_tags:
        info = tags.copy()
        if 'info' in kwargs:
            info.update(kwargs['info'])
        kwargs['info'] = info

    atoms = crystal(symbols, basis=scaled_positions, 
                    cellpar=[a, b, c, alpha, beta, gamma],
                    spacegroup=spacegroup, **kwargs)
    return atoms
    

def read_cif(fileobj, index=-1, store_tags=False, **kwargs):
    """Read Atoms object from CIF file. *index* specifies the data
    block number or name (if string) to return.  

    If *index* is None or a slice object, a list of atoms objects will
    be returned. In the case of *index* is *None* or *slice(None)*,
    only blocks with valid crystal data will be included.

    If *store_tags* is true, the *info* attribute of the returned
    Atoms object will be populated with all tags in the corresponding
    cif data block.

    Keyword arguments are passed on to the Atoms constructor."""

    blocks = parse_cif(fileobj)
    if isinstance(index, str):
        tags = dict(blocks)[index]
        return tags2atoms(tags, **kwargs)
    elif isinstance(index, int):
        name, tags = blocks[index]
        return tags2atoms(tags, **kwargs)
    elif index is None or index == slice(None):
        # Return all CIF blocks with valid crystal data
        images = []
        for name, tags in blocks:
            try:
                atoms = tags2atoms(tags)
                images.append(atoms)
            except KeyError:
                pass
        if not images:
            # No block contained a a valid atoms object
            # Provide an useful error by try converting the first
            # block to atoms
            name, tags = blocks[0]
            tags2atoms(tags)
        return images
    else:
        return [tags2atoms(tags) for name, tags in blocks[index]]


def write_cif(fileobj, images):
    """Write *images* to CIF file."""
    if isinstance(fileobj, str):
        fileobj = paropen(fileobj, 'w')

    if not isinstance(images, (list, tuple)):
        images = [images]

    for i, atoms in enumerate(images):
        fileobj.write('data_image%d\n' % i)

        from numpy import arccos, pi, dot
        from numpy.linalg import norm

        cell = atoms.cell
        a = norm(cell[0])
        b = norm(cell[1])
        c = norm(cell[2])
        alpha = arccos(dot(cell[1], cell[2])/(b*c))*180./pi
        beta = arccos(dot(cell[0], cell[2])/(a*c))*180./pi
        gamma = arccos(dot(cell[0], cell[1])/(a*b))*180./pi

        fileobj.write('_cell_length_a       %g\n' % a)
        fileobj.write('_cell_length_b       %g\n' % b)
        fileobj.write('_cell_length_c       %g\n' % c)
        fileobj.write('_cell_angle_alpha    %g\n' % alpha)
        fileobj.write('_cell_angle_beta     %g\n' % beta)
        fileobj.write('_cell_angle_gamma    %g\n' % gamma)
        fileobj.write('\n')

        if atoms.pbc.all():
            fileobj.write('_symmetry_space_group_name_H-M    %s\n' % 'P 1')
            fileobj.write('_symmetry_int_tables_number       %d\n' % 1)
            fileobj.write('\n')

            fileobj.write('loop_\n')
            fileobj.write('  _symmetry_equiv_pos_as_xyz\n')
            fileobj.write("  'x, y, z'\n")
            fileobj.write('\n')

        fileobj.write('loop_\n')
        fileobj.write('  _atom_site_label\n')
        fileobj.write('  _atom_site_occupancy\n')
        fileobj.write('  _atom_site_fract_x\n')
        fileobj.write('  _atom_site_fract_y\n')
        fileobj.write('  _atom_site_fract_z\n')
        fileobj.write('  _atom_site_thermal_displace_type\n')
        fileobj.write('  _atom_site_B_iso_or_equiv\n')
        fileobj.write('  _atom_site_type_symbol\n')

        scaled = atoms.get_scaled_positions()
        no = {}
        for i, atom in enumerate(atoms):
            symbol = atom.symbol
            if symbol in no:
                no[symbol] += 1
            else:
                no[symbol] = 1
            fileobj.write(
                '  %-8s %6.4f %7.5f  %7.5f  %7.5f  %4s  %6.3f  %s\n'%(
                    '%s%d' % (symbol, no[symbol]), 
                    1.0, 
                    scaled[i][0], 
                    scaled[i][1], 
                    scaled[i][2],
                    'Biso',
                    1.0,
                    symbol))

    



from django.db import models
from composition import Composition

from utils import *

class Author(models.Model):
    name = models.CharField(max_length=100)
    class Meta:
        db_table = 'authors'

    def __str__(self):
        return self.name

class Journal(models.Model):
    code = models.CharField(max_length=10, unique=True)
    full_name = models.TextField(null=True)
    class Meta:
        db_table = 'journals'

    def __str__(self):
        return self.code

class Reference(models.Model):
    authors = models.ManyToManyField(Author, related_name='references',
            null=True)
    journal = models.ForeignKey(Journal, related_name='references', 
            null=True)
    title = models.TextField(null=True)
    volume = models.IntegerField(null=True)
    page_first = models.IntegerField(null=True)
    page_last = models.IntegerField(null=True)
    year = models.IntegerField(null=True)
    volume = models.IntegerField(null=True)
    class Meta:
        db_table = 'publications'

    def __str__(self):
        s = self.title
        if self.authors is not None:
            s += ': %s' % self.authors.all()[0]
        if self.journal is not None:
            s += ': %s' % self.journal
        return s

    @property
    def citation(self):
        auths = ', '.join(str(a) for a in self.authors.all())
        return '{auths} ({year}). {title}. {jrnl}, {vol}, {pb}-{pe}'.format(
                auths=auths,
                year=self.year,
                title=self.title,
                jrnl=self.journal,
                vol=self.volume,
                pb=self.page_first,
                pe=self.page_last)

class Experiment(models.Model):
    energy = models.FloatField(null=True)
    delta_e = models.FloatField(null=True)
    delta_g = models.FloatField(null=True)

    temperature = models.FloatField(null=True)
    pressure = models.FloatField(null=True)
    state = models.CharField(max_length=2)

    composition = models.ForeignKey('Composition', null=True)
    name = models.CharField(max_length=255, null=True)
    dft = models.BooleanField(default=False)

    comment = models.CharField(max_length=255, null=True, blank=True)
    source = models.CharField(max_length=10, blank=True, null=True)
    #reference = models.ForeignKey(Reference, null=True, blank=True)

    class Meta:
        db_table = 'experiments'

    def __str__(self):
        return '%s : %s' % (self.composition, self.delta_e)

    @property
    def comp(self):
        return self.composition.comp

    @property
    def unit_comp(self):
        return self.composition.unit_comp

    @staticmethod
    def create(name, delta_e, d_g, source):
        comp = Composition.get(name)
        expt = Experiment(name=name,
                delta_e=delta_e, d_g=d_g,
                composition=comp,
                source=source)
        return expt

class ICSD(models.Model):
    coll_code = models.IntegerField(primary_key=True)
    composition = models.ForeignKey('Composition', null=True,
            related_name='icsd_obj')

    chemical_name = models.TextField()
    chemical_formula_structural = models.TextField()
    chemical_formula_sum = models.TextField()

    temperature = models.FloatField(default=293)
    pressure = models.FloatField(default=101.325)

    spacegroup_HM = models.CharField(max_length=20)
    spacegroup = models.IntegerField(default=1)
    r_factor = models.FloatField()

    cif = models.TextField()
    reference = models.ForeignKey(Reference, null=True)

    class Meta:
        db_table = 'icsd'

    def __str__(self):
        return str(self.coll_code)

    @staticmethod
    def parse_cif(cif):
        from ase.io.cif import parse_cif
        cifdict = parse_cif(cif)[0][1]

        icsd = ICSD()

        icsd.coll_code =\
                cifdict['_database_code_icsd']
        icsd.chemical_formula_structural =\
                cifdict.get('_chemical_formula_structural', '')
        icsd.chemical_name =\
                cifdict.get('_chemical_name_systematic', '')
        icsd.chemical_formula_sum = \
                cifdict.get('_chemical_formula_sum', '')

        icsd.temperature =\
                cifdict.get('_cell_measurement_temperature', 293)
        icsd.pressure =\
                cifdict.get('_cell_measurement_pressure', 101.325)
        icsd.r_factor =\
                cifdict.get('_refine_ls_r_factor_all',0)
        icsd.spacegroup =\
                cifdict.get('_symmetry_int_tables_number', 1)
        icsd.spacegroup_hm =\
                cifdict.get('_symmetry_space_group_name_h-m', 'P1')
        icsd.save()

        for wyck in zip(
                cifdict['_atom_site_type_symbol'],
                cifdict['_atom_site_wyckoff_symbol'],
                cifdict['_atom_site_symmetry_multiplicity'],
                cifdict['_atom_site_fract_x'],
                cifdict['_atom_site_fract_y'],
                cifdict['_atom_site_fract_z'],
                cifdict['_atom_site_occupancy'],
                cifdict['_atom_site_attached_hydrogens']):
            wyck = list(wyck)
            elt = ''
            charge = ''
            negative = False
            for char in wyck[0]:
                if char.isalpha():
                    elt += char
                elif char == '+':
                    continue
                elif char == '-':
                    negative=True
                else:
                    charge += char
            if negative:
                charge = -1*int(charge)
            else:
                charge =int(charge)

            if isinstance(wyck[6], basestring):
                    wyck[6] = float(wyck[6].split('(')[0])

            wyckobj = WyckoffSite(
                    icsd=icsd,
                    element=elt,
                    charge=charge,
                    symbol=wyck[1],
                    multiplicity=wyck[2],
                    x = wyck[3],
                    y = wyck[4],
                    z = wyck[5],
                    occupancy = wyck[6],
                    attached_hydrogens = wyck[7])
            wyckobj.save()
        return icsd

class WyckoffSite(models.Model):
    icsd = models.ForeignKey(ICSD)
    symbol = models.CharField(max_length=1)

    element = models.CharField(max_length=2)
    charge = models.IntegerField()
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    multiplicity = models.IntegerField(default=1)
    occupancy = models.FloatField(default=1)
    attached_hydrogens = models.IntegerField(default=0)

    class Meta:
        db_table = 'wyckoff_sites'

    def __str__(self):
        return '%s on %s' % (self.element, self.symbol)




import os
from os.path import dirname, abspath
import sys

### Kludge to get the django settings module into the path
sys.path.insert(0, dirname(__file__))
if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = 'dev_qmpy.db.settings'

from models import *
from db import *



from distutils.core import setup

setup(
    name='qmpy',
    version='0.0.1',
    author='S. Kirklin',
    author_email='scott.kirklin@gmail.com',
    packages=['qmpy', 'qmpy.analysis', 'qmpy.io', 'qmpy.models',
        'qmpy.configuration', 'qmpy.utils'],
    scripts=['bin/oqmd', 'bin/runner.py'],
    url='http://pypi.python.org/pypi/qmpy',
    license='LICENSE.txt',
    description='Suite of computational materials science tools',
    long_description=open('README.txt').read(),
    install_requires=[
        "Django >= 1.1.1",
        "Pulp >= 1.4.0",
        "numpy >= 1.6.4"
    ],
)



from django.db import models
import json

from utils import *
from thermopy import *

### things for efficiently indexing the master entries

class Fit(models.Model):
    name = models.CharField(max_length=20, primary_key=True)
    experiments = models.ManyToManyField('Experiment')
    dft = models.ManyToManyField('Calculation')
    fit_elements = models.ManyToManyField('Element')
    fit_hubbards = models.ManyToManyField('Hubbard')
    class Meta:
        db_table = 'fits'

    def __str__(self):
        return self.name

    def formation(self, calculation):
        adjust = 0.0
        for elt, amt in calculation.input.unit_comp.items():
            if self.mu_set.filter(element_id=elt).exists():
                adjust -= amt*self.mu_set.get(element_id=elt).energy
            else:
                adjust -= amt*Element.get(elt).base_mu

        for hub in calculation.hubbards:
            if hub in list(self.fit_hubbards.all()):
                if not self.hubbardmu_set.filter(hubbard=hub).exists():
                    continue
                amt = calculation.input.unit_comp[hub.element_id]
                adjust -= self.hubbardmu_set.get(hubbard=hub).energy*amt
        n = calculation.input.natoms
        form, created = Formation.objects.get_or_create(delta_e=calculation.energy_pa + adjust,
                calculation=calculation, fit=self)
        if created:
            form.save()
        return form

    def compute_new_formations(self):
        from calculation import Calculation
        print 'import successful'
        calcs = Calculation.objects.filter(settings='standard',
                done=True).exclude(formation__fit=self)
        for calc in calcs:
            print calc.entry.name
            form = self.formation(calc)
            form.composition = calc.composition
            form.entry = calc.entry
            form.description = calc.entry.label
            form.save()

    def compute_expt_formations(self):
        from calculation import Calculation
        print 'import successful'
        calcs = Calculation.objects.filter(settings='standard',
                done=True, composition__experiment__dft=False).\
                        exclude(formation__fit=self)
        for calc in calcs.distinct():
            print calc.entry.name
            form = self.formation(calc)
            form.composition = calc.composition
            form.entry = calc.entry
            form.description = calc.entry.label
            form.save()

    def compute_all_formations(self):
        from calculation import Calculation
        print 'import successful'
        calcs = Calculation.objects.filter(settings='standard', done=True)
        print 'Todo:', calcs.count()
        for calc in calcs:
            print calc.entry.name
            form = self.formation(calc)
            form.composition = calc.composition
            form.entry = calc.entry
            form.description = calc.entry.label
            form.save()

    @staticmethod
    def get(fitting):
        fit = Fit.objects.filter(name=fitting)
        if not fit.exists():
            print 'Fit by that name doesn\'t exist. Try one of:'
            for f in Fit.objects.all():
                print '   -', f.name
        return fit[0]

    @staticmethod
    def create(label=None, calculations=None, experiments=None, fit_for=[]):
        data = []
        hub_data = []
        fit, created = Fit.objects.get_or_create(name=label)
        if created:
            fit.save()
        fit.formation_set.all().delete()
        fit.mu_set.all().delete()
        fit.hubbardmu_set.all().delete() # fresh start on mus

        elements = set()  # list of all elements in any compound in fit
        hubbards = set() # list of all hubbards in any calc in fit
        fit_set = set(fit_for)
        expts = {}
        calcs = {}

        for expt in experiments:
            if expt.delta_e is None:
                continue
            #name = expt.composition.name
            if not set(expt.comp.keys()) & fit_set:
                continue

            if not expt.name in expts:
                expts[expt.name] = expt

            elif expt.delta_e < expts[expt.name].delta_e:
                expts[expt.name] = expt

        for calc in calculations:
            if calc.energy_pa is None:
                continue
            if not set(calc.composition.comp.keys()) & fit_set:
                print 'nothing from', calc.composition.comp.keys(), 'in', fit_set
                continue
            name = calc.composition.name
            if not name in calcs:
                calcs[name] = calc
            elif calc.energy_pa < calcs[name].energy_pa:
                calcs[name] = calc

        valid_pairs = set(calcs.keys()) & set(expts.keys())
        for name in valid_pairs:
            fit.experiments.add(expts[name])
            fit.dft.add(calcs[name])
            for elt in name_to_comp(name):
                elements.add(elt)
            if not calcs[name].hubbard:
                data.append(name)
            else:
                hub_data.append(name)
                for hub in calcs[name].hubbards:
                    if hub: 
                        hubbards.add(hub)

        elements = list(elements)
        hubbards = list(hubbards)
        hubbard_elements = [ hub.element.symbol for hub in hubbards ]

        A = []
        b = []
        base_mus = dict( (elt, Element.get(elt).mu) for elt in elements)
        pprint(base_mus)
        for elt, mu in base_mus.items():
            elt = Element.get(elt)
            elt.base_mu = mu
            elt.save()

        for name in data:
            unit_comp = name_to_comp(name, special='unit')
            # remove non-fitting elements
            print name, calcs[name].energy_pa, expts[name].delta_e
            b.append(calcs[name].energy_pa - expts[name].delta_e - sum( base_mus[elt]*amt 
                    for elt, amt in unit_comp.items() if elt not in fit_for ))
            A.append([ unit_comp[elt] for elt in fit_for ])

        A = np.array(A)
        b = np.array(b)
#        if len(A) == 0 and len(b) == 0:
#            print 'Nothing in fit set. What did you do wrong?'
#            exit()

        result = np.linalg.lstsq(A, b)
        element_mus = dict(zip(fit_for, result[0]))

        ### Second fit
        A = []
        b = []
        for name in hub_data:
            print 'hubbard:', name, calcs[name].energy_pa, expts[name].delta_e
            unit_comp = name_to_comp(name, special='unit')
            b.append(calcs[name].energy_pa - expts[name].delta_e -
                    sum( base_mus[elt]*amt
                        for elt, amt in unit_comp.items() 
                        if elt not in fit_for) - 
                    sum( element_mus[elt]*amt
                        for elt, amt in unit_comp.items()
                        if elt in fit_for))
            A.append([ unit_comp[elt] for elt in hubbard_elements ])

        A = np.array(A)
        b = np.array(b)
 #       if len(A) == 0 and len(b) == 0:
 #           print 'Nothing in fit set. What did you do wrong?'
 #           exit()
        result = np.linalg.lstsq(A, b)
        hubbard_mus = dict(zip(hubbards, result[0]))

        pprint(hubbard_mus)
        fit.save()

        for elt, val in element_mus.items():
            elt = Element.get(elt)
            print elt, val
            mu = Mu(element=elt, energy=val, fitting=fit)
            mu.save()
            fit.fit_elements.add(elt)

        for hub, val in hubbard_mus.items():
            print hub, val
            hubmu = HubbardMu(hubbard=hub, energy=val, fitting=fit)
            hubmu.save()
            fit.fit_hubbards.add(hub)

        fit.save()
        return fit

    @property
    def mu_dict(self):
        mdict = {}
        for mu in self.mu_set.all():
            mdict[mu.element.symbol] = mu.energy
        return mdict

    @property
    def hub_dict(self):
        hdict = defaultdict(float)
        for hub in self.hubbardmu_set.all():
            hdict[hub.element.symbol] = hub.energy
        return hdict
            
class HubbardMu(models.Model):
    hubbard = models.ForeignKey('Hubbard')
    energy = models.FloatField()
    fitting = models.ForeignKey(Fit)

    class Meta:
        db_table = 'hubbard_mus'

    def __str__(self):
        return '%s=%s' % (self.hubbard.element, self.energy)

class Mu(models.Model):
    element = models.ForeignKey('Element',
            related_name='mus')
    energy = models.FloatField()
    fitting = models.ForeignKey(Fit)

    class Meta:
        db_table = 'mus'

    def __str__(self):
        return '%s=%s' % (self.element, self.energy)


class Formation(models.Model):
    entry = models.ForeignKey('Entry', null=True, blank=True)
    composition = models.ForeignKey('Composition', null=True, blank=True)
    calculation = models.ForeignKey('Calculation', null=True, blank=True)
    description = models.CharField(max_length=20, null=True, blank=True)
    notes = models.CharField(max_length=255, null=True, blank=True)
    fit = models.ForeignKey('Fit', null=True)
    hull_distance = models.FloatField(null=True)
    delta_e = models.FloatField(null=True)
    tielines = models.ManyToManyField('self')

    phase = models.TextField(null=True, blank=True)

    class Meta:
        db_table = 'formations'

    def __str__(self):
        return '%s : %s' % (self.composition, self.delta_e)

    @staticmethod
    def get(calculation, fit='standard'):
        fit = Fit.get(fit)
        form = fit.formation(calculation)
        form.composition = calculation.composition
        form.entry = calculation.entry
        form.description = form.entry.label
        form.save()
        return form.delta_e

    @property
    def phase(self):
        return Phase(energy=self.delta_e, 
                composition=self.composition.comp, 
                description=self.id,
                #description=self.calculation.entry.label,
                normalize_energy=False)

    @property
    def pop_string(self):
        template='''$structure: {comment}
CREATE_NEW_EQUILIBRIUM {id},1
SET-CONDITION P=101325,T=298.15,{composition}
EXPERIMENT H={energy}:1000'''
        comp = self.composition.comp
        composition = ', '.join([ 'N(%s)=%s' % (elt, amt) for 
            elt, amt in comp.items() ])
        energy = 96485.3*self.delta_e*sum(comp.values())
        return template.format(
                comment=self.entry.label,
                composition=composition, 
                energy=energy,
                id=self.id)


    def find_tie_lines(self):
        space = PhaseSpace()
        for formation in Formation.objects.get(fit=self.fit):
            break

class Equilibria(models.Model):
    fit = models.ForeignKey('Fit')
    members = models.ManyToManyField('Formation')
